import enum
import typing
import logging

import attr
import zigpy.zdo.types
import zigpy_znp.types as t


LOGGER = logging.getLogger(__name__)


@attr.s
class BindEntry(t.Struct):
    """Bind table entry."""

    Src = attr.ib(type=t.EUI64, converter=t.Struct.converter(t.EUI64))
    SrcEp = attr.ib(type=t.uint8_t, converter=t.uint8_t)
    ClusterId = attr.ib(type=t.ClusterId, converter=t.ClusterId)
    DstAddr = attr.ib(
        type=zigpy.zdo.types.MultiAddress,
        converter=t.Struct.converter(zigpy.zdo.types.MultiAddress),
    )


class CommandType(t.enum_uint8):
    """Command Type."""

    POLL = 0
    SREQ = 1  # A synchronous request that requires an immediate response
    AREQ = 2  # An asynchronous request
    SRSP = 3  # A synchronous response. This type of command is only sent in response to
    # a SREQ command
    RESERVED_4 = 4
    RESERVED_5 = 5
    RESERVED_6 = 6
    RESERVED_7 = 7


class ErrorCode(t.MissingEnumMixin, t.enum_uint8):
    """Error code."""

    INVALID_SUBSYSTEM = 0x01
    INVALID_COMMAND_ID = 0x02
    INVALID_PARAMETER = 0x03
    INVALID_LENGTH = 0x04


class Subsystem(t.enum_uint8):
    """Command subsystem."""

    RPCError = 0x00
    SYS = 0x01
    MAC = 0x02
    NWK = 0x03
    AF = 0x04
    ZDO = 0x05
    SAPI = 0x06
    UTIL = 0x07
    DEBUG = 0x08
    APP = 0x09
    RESERVED_10 = 0x0A
    RESERVED_11 = 0x0B
    RESERVED_12 = 0x0C
    RESERVED_13 = 0x0D
    RESERVED_14 = 0x0E
    APPConfig = 0x0F
    RESERVED_16 = 0x10
    RESERVED_17 = 0x11
    RESERVED_18 = 0x12
    RESERVED_19 = 0x13
    RESERVED_20 = 0x14
    ZGP = 0x15
    RESERVED_22 = 0x16
    RESERVED_23 = 0x17
    RESERVED_24 = 0x18
    RESERVED_25 = 0x19
    RESERVED_26 = 0x1A
    RESERVED_27 = 0x1B
    RESERVED_28 = 0x1C
    RESERVED_29 = 0x1D
    RESERVED_30 = 0x1E
    RESERVED_31 = 0x1F


class CallbackSubsystem(t.enum_uint16):
    """Subscribe/unsubscribe subsystem callbacks."""

    MT_SYS = Subsystem.SYS << 8
    MC_MAC = Subsystem.MAC << 8
    MT_NWK = Subsystem.NWK << 8
    MT_AF = Subsystem.AF << 8
    MT_ZDO = Subsystem.ZDO << 8
    MT_SAPI = Subsystem.SAPI << 8
    MT_UTIL = Subsystem.UTIL << 8
    MT_DEBUG = Subsystem.DEBUG << 8
    MT_APP = Subsystem.APP << 8
    MT_APPConfig = Subsystem.APPConfig << 8
    MT_ZGP = Subsystem.ZGP << 8
    ALL = 0xFFFF


class CommandHeader(t.uint16_t):
    """CommandHeader class."""

    def __new__(
        cls, value: int = 0x0000, *, id=None, subsystem=None, type=None
    ) -> "CommandHeader":
        instance = super().__new__(cls, value)

        if id is not None:
            instance = instance.with_id(id)

        if subsystem is not None:
            instance = instance.with_subsystem(subsystem)

        if type is not None:
            instance = instance.with_type(type)

        return instance

    @property
    def cmd0(self) -> t.uint8_t:
        return t.uint8_t(self & 0x00FF)

    @property
    def id(self) -> t.uint8_t:
        """Return CommandHeader id."""
        return t.uint8_t(self >> 8)

    def with_id(self, value: int) -> "CommandHeader":
        """command ID setter."""
        return type(self)(self & 0x00FF | (value & 0xFF) << 8)

    @property
    def subsystem(self) -> Subsystem:
        """Return subsystem of the command."""
        return Subsystem(self.cmd0 & 0x1F)

    def with_subsystem(self, value: Subsystem) -> "CommandHeader":
        return type(self)(self & 0xFFE0 | value & 0x1F)

    @property
    def type(self) -> CommandType:
        """Return command type."""
        return CommandType(self.cmd0 >> 5)

    def with_type(self, value) -> "CommandHeader":
        return type(self)(self & 0xFF1F | (value & 0x07) << 5)

    def __str__(self) -> str:
        return (
            f"{type(self).__name__}("
            f"id=0x{self.id:02X}, "
            f"subsystem={self.subsystem!s}, "
            f"type={self.type!s}"
            ")"
        )

    __repr__ = __str__


@attr.s
class CommandDef:
    command_type: CommandType = attr.ib()
    command_id: int = attr.ib()
    req_schema: typing.Optional[t.Schema] = attr.ib(
        factory=lambda *v: None if not v else t.Schema(*v)
    )
    rsp_schema: typing.Optional[t.Schema] = attr.ib(
        factory=lambda *v: None if not v else t.Schema(*v)
    )


class CommandsMeta(type):
    """
    Metaclass that creates `Command` subclasses out of the `CommandDef` definitions
    """

    def __new__(cls, name: str, bases, classdict, *, subsystem):
        # Ignore CommandsBase
        if not bases:
            return type.__new__(cls, name, bases, classdict)

        classdict["_commands"] = []

        for command, definition in classdict.items():
            if not isinstance(definition, CommandDef):
                continue

            # We manually create the qualname to match the final object structure
            qualname = classdict["__qualname__"] + "." + command

            # The commands class is dynamically created from the definition
            helper_class_dict = {
                "definition": definition,
                "type": definition.command_type,
                "subsystem": subsystem,
                "__qualname__": qualname,
                "Req": None,
                "Rsp": None,
                "Callback": None,
            }

            header = (
                CommandHeader()
                .with_id(definition.command_id)
                .with_type(definition.command_type)
                .with_subsystem(subsystem)
            )

            rsp_header = header

            # TODO: explore __set_name__

            if definition.req_schema is not None:
                # AREQ doesn't necessarily mean it's a callback
                # Some requests don't have any response at all
                if definition.command_type == CommandType.AREQ:

                    class Req(CommandBase, header=header, schema=definition.req_schema):
                        pass

                    Req.__qualname__ = qualname + ".Req"
                    Req.Req = Req
                    Req.Rsp = None
                    Req.Callback = None
                    helper_class_dict["Req"] = Req
                else:
                    req_header = header
                    rsp_header = CommandHeader(0x0040 + req_header)

                    class Req(
                        CommandBase, header=req_header, schema=definition.req_schema
                    ):
                        pass

                    class Rsp(
                        CommandBase, header=rsp_header, schema=definition.rsp_schema
                    ):
                        pass

                    Req.__qualname__ = qualname + ".Req"
                    Req.Req = Req
                    Req.Rsp = Rsp
                    Req.Callback = None
                    helper_class_dict["Req"] = Req

                    Rsp.__qualname__ = qualname + ".Rsp"
                    Rsp.Req = Rsp
                    Rsp.Req = Req
                    Rsp.Callback = None
                    helper_class_dict["Rsp"] = Rsp
            else:
                assert definition.rsp_schema is not None, definition

                if definition.command_type == CommandType.AREQ:
                    # If there is no request schema, this is a callback
                    class Callback(
                        CommandBase, header=header, schema=definition.rsp_schema
                    ):
                        pass

                    Callback.__qualname__ = qualname + ".Callback"
                    Callback.Req = None
                    Callback.Rsp = None
                    Callback.Callback = Callback
                    helper_class_dict["Callback"] = Callback
                elif definition.command_type == CommandType.SRSP:
                    # XXX: This is the only command like this
                    #      everything else should be an error!
                    if header != CommandHeader(
                        subsystem=Subsystem.RPCError, id=0x00, type=CommandType.SRSP
                    ):
                        raise RuntimeError(
                            f"Invalid command definition {command} = {definition}"
                        )  # pragma: no cover

                    # If there is no request, this is a just a response
                    class Rsp(CommandBase, header=header, schema=definition.rsp_schema):
                        pass

                    Rsp.__qualname__ = qualname + ".Rsp"
                    Rsp.Req = None
                    Rsp.Rsp = Rsp
                    Rsp.Callback = None
                    helper_class_dict["Rsp"] = Rsp
                else:
                    raise RuntimeError(
                        f"Invalid command definition {command} = {definition}"
                    )  # pragma: no cover

            classdict[command] = type(command, (), helper_class_dict)
            classdict["_commands"].append(classdict[command])

        return type.__new__(cls, name, bases, classdict)

    def __iter__(self):
        return iter(self._commands)


class CommandsBase(metaclass=CommandsMeta, subsystem=None):
    pass


class CommandBase:
    Req = None
    Rsp = None
    Callback = None

    def __init_subclass__(cls, *, header, schema):
        super().__init_subclass__()
        cls.header = header
        cls.schema = schema

    def __init__(self, *, partial=False, **params):
        all_params = {param.name for param in self.schema.parameters}
        given_params = set(params.keys())

        if given_params - all_params:
            raise KeyError(
                f"Unexpected parameters: {given_params - all_params}. "
                f"Expected one of {all_params - given_params}"
            )
        elif not partial and all_params - given_params:
            raise KeyError(f"Missing parameters: {all_params - given_params}")

        bound_params = {}

        for param in self.schema.parameters:
            if partial and params.get(param.name) is None:
                bound_params[param.name] = (param, None)
                continue

            value = params[param.name]

            if not isinstance(value, param.type):
                # fmt: off
                is_coercible_type = [
                    isinstance(value, int)
                    and issubclass(param.type, int)
                    and not issubclass(param.type, enum.Enum),

                    isinstance(value, bytes)
                    and issubclass(param.type, (t.ShortBytes, t.LongBytes)),

                    isinstance(value, list) and issubclass(param.type, list),
                    isinstance(value, bool) and issubclass(param.type, t.Bool),
                ]
                # fmt: on

                if any(is_coercible_type):
                    value = param.type(value)
                elif (
                    type(value) is zigpy.zdo.types.SimpleDescriptor
                    and param.type is zigpy.zdo.types.SizePrefixedSimpleDescriptor
                ):
                    data = value.serialize()
                    value, _ = param.type.deserialize(bytes([len(data)]) + data)
                else:
                    raise ValueError(
                        f"In {type(self)}, param {param.name} is "
                        f"type {param.type}, got {type(value)}"
                    )

            try:
                # XXX: Break early if a type cannot be serialized
                value.serialize()
            except Exception as e:
                raise ValueError(
                    f"Invalid parameter value: {param.name}={value!r}"
                ) from e

            bound_params[param.name] = (param, value)

        super().__setattr__("_bound_params", bound_params)

    def to_frame(self):
        from zigpy_znp.frames import GeneralFrame

        missing_params = {p.name for p, v in self._bound_params.values() if v is None}

        if missing_params:
            raise ValueError(
                f"Cannot serialize a partial frame: missing {missing_params}"
            )

        data = b"".join([v.serialize() for p, v in self._bound_params.values()])

        return GeneralFrame(self.header, data)

    @classmethod
    def from_frame(cls, frame, *, ignore_unparsed=False) -> "CommandBase":
        if frame.header != cls.header:
            raise ValueError(
                f"Wrong frame header in {cls}: {cls.header} != {frame.header}"
            )

        data = frame.data
        params = {}

        for param in cls.schema.parameters:
            params[param.name], data = param.type.deserialize(data)

        if data:
            if ignore_unparsed:
                LOGGER.warning(
                    f"Unparsed data remains in {cls} at the end of the frame: {data!r}"
                )
            else:
                raise ValueError(
                    f"Unparsed data remains in {cls} at the end of the frame: {data!r}"
                )

        return cls(**params)

    def matches(self, other: "CommandBase") -> bool:
        if type(self) is not type(other):
            return False

        assert self.header == other.header

        param_pairs = zip(self._bound_params.values(), other._bound_params.values())

        for (
            (expected_param, expected_value),
            (actual_param, actual_value),
        ) in param_pairs:
            assert expected_param == actual_param

            # Only non-None bound params are considered
            if expected_value is not None and expected_value != actual_value:
                return False

        return True

    def __eq__(self, other):
        return type(self) is type(other) and self._bound_params == other._bound_params

    def __hash__(self):
        params = tuple(self._bound_params.items())
        return hash((type(self), self.header, self.schema, params))

    def __getattr__(self, key):
        if key not in self._bound_params:
            raise AttributeError(f"{self} has no attribute {key!r}")

        param, value = self._bound_params[key]
        return value

    def __setattr__(self, key, value):
        raise RuntimeError("Command instances are immutable")

    def __delattr__(self, key):
        raise RuntimeError("Command instances are immutable")

    def __repr__(self):
        params = [f"{p.name}={v!r}" for p, v in self._bound_params.values()]

        return f'{self.__class__.__qualname__}({", ".join(params)})'

    __str__ = __repr__


class DeviceState(t.MissingEnumMixin, t.enum_uint8):
    """Indicated device state."""

    InitializedNotStarted = 0x00
    InitializedNotConnected = 0x01
    DiscoveringPANs = 0x02
    Joining = 0x03
    ReJoining = 0x04
    JoinedNotAuthenticated = 0x05
    JoinedAsEndDevice = 0x06
    JoinedAsRouter = 0x07
    StartingAsCoordinator = 0x08
    StartedAsCoordinator = 0x09
    LostParent = 0x0A


class InterPanCommand(t.enum_uint8):
    InterPanClr = 0x00
    InterPanSet = 0x01
    InterPanReg = 0x02
    InterPanChk = 0x03


class MTCapabilities(t.enum_flag_uint16):
    CAP_SYS = 1 << 0
    CAP_MAC = 1 << 1
    CAP_NWK = 1 << 2
    CAP_AF = 1 << 3
    CAP_ZDO = 1 << 4
    CAP_SAPI = 1 << 5
    CAP_UTIL = 1 << 6
    CAP_DEBUG = 1 << 7
    CAP_APP = 1 << 8
    CAP_GP = 1 << 9
    CAP_APP_CNF = 1 << 10
    CAP_UNK12 = 1 << 11
    CAP_ZOAD = 1 << 12
    CAP_UNK14 = 1 << 13
    CAP_UNK15 = 1 << 14
    CAP_UNK16 = 1 << 15


@attr.s
class Network(t.Struct):
    PanId = attr.ib(type=t.PanId, converter=t.Struct.converter(t.PanId))
    Channel = attr.ib(type=t.uint8_t, converter=t.uint8_t)
    StackProfileVersion = attr.ib(type=t.uint8_t, converter=t.uint8_t)
    BeaconOrderSuperframe = attr.ib(type=t.uint8_t, converter=t.uint8_t)
    PermitJoining = attr.ib(type=t.uint8_t, converter=t.uint8_t)


STATUS_SCHEMA = t.Schema(
    (t.Param("Status", t.Status, "Status is either Success (0) or Failure (1)"),)
)



import asyncio
import logging
import typing

import serial
import serial.tools
import serial_asyncio

from serial.tools.list_ports import comports as list_com_ports

import zigpy_znp.config as conf
import zigpy_znp.frames as frames

from zigpy_znp.types import Bytes
from zigpy_znp.exceptions import InvalidFrame

LOGGER = logging.getLogger(__name__)


class BufferTooShort(Exception):
    pass


class ZnpMtProtocol(asyncio.Protocol):
    def __init__(self, api):
        self._buffer = bytearray()
        self._api = api
        self._transport = None
        self._connected_event = asyncio.Event()

    @property
    def transport(self) -> typing.Optional[serial_asyncio.SerialTransport]:
        """Return current transport."""
        return self._transport

    def close(self) -> None:
        """Closes the port."""
        self._buffer.clear()
        self.transport.close()

    def connection_lost(self, exc: typing.Optional[Exception]) -> None:
        """Connection lost."""

        self._buffer.clear()

        if exc is not None:
            LOGGER.warning(
                "Lost connection to %s", self._transport.serial.name, exc_info=exc
            )

        self._api.connection_lost(exc)

        LOGGER.debug("Closing %s serial port", self._transport.serial.name)

    def connection_made(self, transport: serial_asyncio.SerialTransport) -> None:
        """Opened serial port."""
        self._transport = transport
        LOGGER.debug("Opened %s serial port", transport.serial.name)

        self._connected_event.set()

    def data_received(self, data: bytes) -> None:
        """Callback when data is received."""
        self._buffer += data

        LOGGER.trace("Received data: %s", Bytes.__repr__(data))

        for frame in self._extract_frames():
            LOGGER.trace("Parsed frame: %s", frame)

            try:
                self._api.frame_received(frame.payload)
            except Exception as e:
                LOGGER.error(
                    "Received an exception while passing frame to API", exc_info=e
                )

    def send(self, payload: frames.GeneralFrame) -> None:
        """Sends data taking care of framing."""
        data = frames.TransportFrame(payload).serialize()
        LOGGER.trace("Sending data: %s", Bytes.__repr__(data))
        self.transport.write(data)

    def _extract_frames(self) -> typing.Iterator[frames.TransportFrame]:
        """Extracts frames from the buffer until it is exhausted."""
        while True:
            try:
                yield self._extract_frame()
            except BufferTooShort:
                # If the buffer is too short, there is nothing more we can do
                break
            except InvalidFrame:
                # If the buffer contains invalid data, drop it until we find the SoF
                sof_index = self._buffer.find(frames.TransportFrame.SOF, 1)

                if sof_index < 0:
                    # If we don't have a SoF in the buffer, drop everything
                    self._buffer.clear()
                else:
                    del self._buffer[:sof_index]

    def _extract_frame(self) -> frames.TransportFrame:
        """Extracts a single frame from the buffer."""

        # The shortest possible frame is 5 bytes long
        if len(self._buffer) < 5:
            raise BufferTooShort()

        # The buffer must start with a SoF
        if self._buffer[0] != frames.TransportFrame.SOF:
            raise InvalidFrame()

        length = self._buffer[1]

        # If the packet length field exceeds 250, our packet is not valid
        if length > 250:
            raise InvalidFrame()

        # Don't bother deserializing anything if the packet is too short
        # [SoF:1] [Length:1] [Command:2] [Data:(Length)] [FCS:1]
        if len(self._buffer) < length + 5:
            raise BufferTooShort()

        # At this point we should have a complete frame
        # If not, deserialization will fail and the error will propapate up
        frame, rest = frames.TransportFrame.deserialize(self._buffer)

        # If we get this far then we have a valid frame. Update the buffer.
        del self._buffer[: len(self._buffer) - len(rest)]

        return frame

    def __repr__(self):
        return f"<{type(self).__name__} for {self._api}>"


def guess_port() -> str:
    """Picks the first USB port with a Texas Instruments vendor ID."""
    candidates = []

    for port in list_com_ports(include_links=True):
        # Add only TI devices
        if port.vid == 0x0451:
            candidates.append(port)

    if not candidates:
        raise RuntimeError("Could not auto detect any TI ports")

    # Is there no better heuristic than picking the first TI device?
    candidates.sort(key=lambda p: p.location)
    device = candidates[0].device

    if len(candidates) > 1:
        LOGGER.warning(
            "Found multiple Texas Instruments devices: %s",
            [c.__dict__ for c in candidates],
        )
        LOGGER.warning("Picking the first one: %s", device)

    return device


async def connect(config: conf.ConfigType, api) -> ZnpMtProtocol:
    loop = asyncio.get_running_loop()

    port = config[conf.CONF_DEVICE_PATH]
    baudrate = config[conf.CONF_DEVICE_BAUDRATE]
    flow_control = config[conf.CONF_DEVICE_FLOW_CONTROL]

    if port == "auto":
        port = guess_port()

    xonxoff, rtscts = {
        None: (False, False),
        "hardware": (False, True),
        "software": (True, False),
    }[flow_control]

    LOGGER.debug("Connecting to %s at %s baud", port, baudrate)

    transport, protocol = await serial_asyncio.create_serial_connection(
        loop,
        lambda: ZnpMtProtocol(api),
        url=port,
        baudrate=baudrate,
        parity=serial.PARITY_NONE,
        stopbits=serial.STOPBITS_ONE,
        xonxoff=xonxoff,
        rtscts=rtscts,
    )

    await protocol._connected_event.wait()

    LOGGER.debug("Connected to %s at %s baud", port, baudrate)

    return protocol



"""ZDO Interface

This interface allows the tester to issue commands to the ZDO layer in the target
and receive responses. Each of these messages has a corresponding message that is
returned by the target. The response message only indicates that the command message
was received and executed. The result of the command execution will be conveyed to
the tester via a callback message interface"""

import typing

import zigpy.zdo.types
import zigpy_znp.types as t


class SecurityEntry(t.FixedList, item_type=t.uint8_t, length=5):
    pass


class StartupState(t.enum_uint8):
    RestoredNetworkState = 0x00
    NewNetworkState = 0x01
    NotStarted = 0x02


class RouteDiscoveryOptions(t.enum_uint8):
    Suppress = 0x00
    Enable = 0x01
    Force = 0x02


class MACCapabilities(t.enum_flag_uint8):
    PANCoordinator = 1 << 0
    Router = 1 << 1
    MainsPowered = 1 << 2
    RXWhenIdle = 1 << 3
    Reserved5 = 1 << 4
    Reserved6 = 1 << 5
    SecurityCapable = 1 << 6
    AllocateShortAddrDuringAssocNeeded = 1 << 7


class LeaveOptions(t.enum_flag_uint8):
    NONE = 0
    Rejoin = 1 << 0
    RemoveChildren = 1 << 1


class NetworkList(t.LVList, item_type=t.Network, length_type=t.uint8_t):
    pass


class EndpointList(t.LVList, item_type=t.uint8_t, length_type=t.uint8_t):
    pass


class GroupIdList(t.LVList, item_type=t.GroupId, length_type=t.uint8_t):
    pass


class BindEntryList(t.LVList, item_type=t.BindEntry, length_type=t.uint8_t):
    pass


class BeaconList(t.LVList, item_type=t.Beacon, length_type=t.uint8_t):
    pass


class NullableNodeDescriptor(zigpy.zdo.types.NodeDescriptor):
    @classmethod
    def deserialize(cls, data: bytes) -> typing.Tuple["NullableNodeDescriptor", bytes]:
        if data == b"\x00":
            return cls(), b""

        return super().deserialize(data)

    def serialize(self) -> bytes:
        # Special case when the node descriptor is completely empty
        if all(getattr(self, field) is None for field, _ in self._fields):
            return b"\x00"

        return super().serialize()


class ZDO(t.CommandsBase, subsystem=t.Subsystem.ZDO):
    # send a “Network Address Request”. This message sends a broadcast message looking
    # for a 16 bit address with a known 64 bit IEEE address. You must subscribe to
    # “ZDO Network Address Response” to receive the response to this message
    NwkAddrReq = t.CommandDef(
        t.CommandType.SREQ,
        0x00,
        req_schema=t.Schema(
            (
                t.Param(
                    "IEEE",
                    t.EUI64,
                    "Extended address of the device requesting association",
                ),
                t.Param(
                    "RequestType",
                    t.uint8_t,
                    "0x00 -- single device request, 0x01 -- Extended",
                ),
                t.Param(
                    "StartIndex", t.uint8_t, "Starting index into the list of children"
                ),
            )
        ),
        rsp_schema=t.STATUS_SCHEMA,
    )

    # request a device’s IEEE 64-bit address
    IEEEAddrReq = t.CommandDef(
        t.CommandType.SREQ,
        0x01,
        req_schema=t.Schema(
            (
                t.Param("NWK", t.NWK, "Short address of the device"),
                t.Param(
                    "RequestType",
                    t.uint8_t,
                    "0x00 -- single device request, 0x01 -- Extended",
                ),
                t.Param(
                    "StartIndex", t.uint8_t, "Starting index into the list of children"
                ),
            )
        ),
        rsp_schema=t.STATUS_SCHEMA,
    )

    # inquire about the Node Descriptor information of the destination device
    NodeDescReq = t.CommandDef(
        t.CommandType.SREQ,
        0x02,
        req_schema=t.Schema(
            (
                t.Param(
                    "DstAddr",
                    t.NWK,
                    "Short address of the device generating the inquiry",
                ),
                t.Param(
                    "NWKAddrOfInterest",
                    t.NWK,
                    "Short address of the device being queried",
                ),
            )
        ),
        rsp_schema=t.STATUS_SCHEMA,
    )

    # inquire about the Power Descriptor information of the destination device
    PowerDescReq = t.CommandDef(
        t.CommandType.SREQ,
        0x03,
        req_schema=t.Schema(
            (
                t.Param(
                    "DstAddr",
                    t.NWK,
                    "Short address of the device generating the inquiry",
                ),
                t.Param(
                    "NWKAddrOfInterest",
                    t.NWK,
                    "Short address of the device being queried",
                ),
            )
        ),
        rsp_schema=t.STATUS_SCHEMA,
    )

    # inquire as to the Simple Descriptor of the destination device’s Endpoint
    SimpleDescReq = t.CommandDef(
        t.CommandType.SREQ,
        0x04,
        req_schema=t.Schema(
            (
                t.Param(
                    "DstAddr",
                    t.NWK,
                    "Short address of the device generating the inquiry",
                ),
                t.Param(
                    "NWKAddrOfInterest",
                    t.NWK,
                    "Short address of the device being queried",
                ),
                t.Param("Endpoint", t.uint8_t, "application endpoint the data is from"),
            )
        ),
        rsp_schema=t.STATUS_SCHEMA,
    )

    # request a list of active endpoint from the destination device
    ActiveEpReq = t.CommandDef(
        t.CommandType.SREQ,
        0x05,
        req_schema=t.Schema(
            (
                t.Param(
                    "DstAddr",
                    t.NWK,
                    "Short address of the device generating the inquiry",
                ),
                t.Param(
                    "NWKAddrOfInterest",
                    t.NWK,
                    "Short address of the device being queried",
                ),
            )
        ),
        rsp_schema=t.STATUS_SCHEMA,
    )

    # request the device match descriptor
    MatchDescReq = t.CommandDef(
        t.CommandType.SREQ,
        0x06,
        req_schema=t.Schema(
            (
                t.Param(
                    "DstAddr",
                    t.NWK,
                    "Short address of the device generating the inquiry",
                ),
                t.Param(
                    "NWKAddrOfInterest",
                    t.NWK,
                    "Short address of the device being queried",
                ),
                t.Param("ProfileId", t.uint16_t, "profile id of the device"),
                t.Param("InputClusters", t.ClusterIdList, "Input cluster id list"),
                t.Param("OutputClusters", t.ClusterIdList, "Output cluster id list"),
            )
        ),
        rsp_schema=t.STATUS_SCHEMA,
    )

    # request for the destination device’s complex descriptor
    ComplexDescReq = t.CommandDef(
        t.CommandType.SREQ,
        0x07,
        req_schema=t.Schema(
            (
                t.Param(
                    "DstAddr",
                    t.NWK,
                    "Short address of the device generating the inquiry",
                ),
                t.Param(
                    "NWKAddrOfInterest",
                    t.NWK,
                    "Short address of the device being queried",
                ),
            )
        ),
        rsp_schema=t.STATUS_SCHEMA,
    )

    # request for the destination device’s user descriptor
    UserDescReq = t.CommandDef(
        t.CommandType.SREQ,
        0x08,
        req_schema=t.Schema(
            (
                t.Param(
                    "DstAddr",
                    t.NWK,
                    "Short address of the device generating the inquiry",
                ),
                t.Param(
                    "NWKAddrOfInterest",
                    t.NWK,
                    "Short address of the device being queried",
                ),
            )
        ),
        rsp_schema=t.STATUS_SCHEMA,
    )

    # This command will cause the CC2480 device to issue an “End device announce”
    # broadcast packet to the network. This is typically used by an end-device to
    # announce itself to the network
    EndDeviceAnnce = t.CommandDef(
        t.CommandType.SREQ,
        0x0A,
        req_schema=t.Schema(
            (
                t.Param("NWK", t.NWK, "Short address of the device"),
                t.Param(
                    "IEEE",
                    t.EUI64,
                    "Extended address of the device generating the request",
                ),
                t.Param("Capabilities", t.uint8_t, "MAC Capabilities"),
            )
        ),
        rsp_schema=t.STATUS_SCHEMA,
    )

    # write a User Descriptor value to the targeted device
    UserDescSet = t.CommandDef(
        t.CommandType.SREQ,
        0x0B,
        req_schema=t.Schema(
            (
                t.Param(
                    "DstAddr",
                    t.NWK,
                    "network address of the device generating the set request",
                ),
                t.Param(
                    "NWK", t.NWK, "NWK address of the destination device being queried"
                ),
                t.Param("UserDescriptor", t.ShortBytes, "User descriptor array"),
            )
        ),
        rsp_schema=t.STATUS_SCHEMA,
    )

    # discover the location of a particular system server or servers as indicated by
    # the ServerMask parameter. The destination addressing on this request is
    # ‘broadcast to all RxOnWhenIdle devices’
    ServerDiscReq = t.CommandDef(
        t.CommandType.SREQ,
        0x0C,
        req_schema=t.Schema(
            (
                t.Param(
                    "ServerMask", t.uint16_t, "system server capabilities of the device"
                ),
            )
        ),
        rsp_schema=t.STATUS_SCHEMA,
    )

    # request an End Device Bind with the destination device
    EndDeviceBindReq = t.CommandDef(
        t.CommandType.SREQ,
        0x20,
        req_schema=t.Schema(
            (
                t.Param(
                    "DstAddr",
                    t.NWK,
                    "Short address of the device generating the request",
                ),
                t.Param(
                    "LocalCoordinator",
                    t.NWK,
                    (
                        "local coordinator’s short address. In the case of source "
                        "binding, it’s the short address of the source address"
                    ),
                ),
                t.Param("IEEE", t.EUI64, "Local coordinator's IEEE address"),
                t.Param("Endpoint", t.uint8_t, "device's endpoint"),
                t.Param("ProfileId", t.uint16_t, "profile id of the device"),
                t.Param("InputClusters", t.ClusterIdList, "Input cluster id list"),
                t.Param("OutputClusters", t.ClusterIdList, "Output cluster id list"),
            )
        ),
        rsp_schema=t.STATUS_SCHEMA,
    )

    # request a Bind
    BindReq = t.CommandDef(
        t.CommandType.SREQ,
        0x21,
        req_schema=t.Schema(
            (
                t.Param("Dst", t.NWK, "Short address of the destination device"),
                t.Param("Src", t.EUI64, "Binding source IEEE address"),
                t.Param("SrcEndpoint", t.uint8_t, "binding source endpoint"),
                t.Param("ClusterId", t.ClusterId, "Cluster id to match in messages"),
                t.Param(
                    "Address", zigpy.zdo.types.MultiAddress, "Binding address/endpoint"
                ),
            )
        ),
        rsp_schema=t.STATUS_SCHEMA,
    )

    # request a UnBind
    UnBindReq = t.CommandDef(
        t.CommandType.SREQ,
        0x22,
        req_schema=t.Schema(
            (
                t.Param("Dst", t.NWK, "Short address of the destination device"),
                t.Param("Src", t.EUI64, "Binding source IEEE address"),
                t.Param("SrcEndpoint", t.uint8_t, "binding source endpoint"),
                t.Param("ClusterId", t.ClusterId, "Cluster id to match in messages"),
                t.Param(
                    "Address",
                    t.AddrModeAddress,
                    "Binding address mode destination address",
                ),
                t.Param("DstEndpoint", t.uint8_t, "Binding destination endpoint"),
            )
        ),
        rsp_schema=t.STATUS_SCHEMA,
    )

    # request the destination device to perform a network discovery
    MgmtNwkDiscReq = t.CommandDef(
        t.CommandType.SREQ,
        0x30,
        req_schema=t.Schema(
            (
                t.Param("Dst", t.NWK, "Short address of the destination device"),
                t.Param("Channels", t.Channels, "Bitmask of channels to scan"),
                t.Param("ScanDuration", t.uint8_t, "Scanning time"),
                t.Param(
                    "StartIndex",
                    t.uint8_t,
                    "Specifies where to start in the response array",
                ),
            )
        ),
        rsp_schema=t.STATUS_SCHEMA,
    )

    # request the destination device to perform a LQI query of other devices
    # in the network
    MgmtLqiReq = t.CommandDef(
        t.CommandType.SREQ,
        0x31,
        req_schema=t.Schema(
            (
                t.Param("Dst", t.NWK, "Short address of the destination device"),
                t.Param(
                    "StartIndex",
                    t.uint8_t,
                    "Specifies where to start in the response array",
                ),
            )
        ),
        rsp_schema=t.STATUS_SCHEMA,
    )

    # request the Routing Table of the destination device
    MgmtRtgReq = t.CommandDef(
        t.CommandType.SREQ,
        0x32,
        req_schema=t.Schema(
            (
                t.Param("Dst", t.NWK, "Short address of the destination device"),
                t.Param(
                    "StartIndex",
                    t.uint8_t,
                    "Specifies where to start in the response array",
                ),
            )
        ),
        rsp_schema=t.STATUS_SCHEMA,
    )

    # request the Binding Table of the destination device
    MgmtBindReq = t.CommandDef(
        t.CommandType.SREQ,
        0x33,
        req_schema=t.Schema(
            (
                t.Param("Dst", t.NWK, "Short address of the destination device"),
                t.Param(
                    "StartIndex",
                    t.uint8_t,
                    "Specifies where to start in the response array",
                ),
            )
        ),
        rsp_schema=t.STATUS_SCHEMA,
    )

    # request a Management Leave Request for the target device
    MgmtLeaveReq = t.CommandDef(
        t.CommandType.SREQ,
        0x34,
        req_schema=t.Schema(
            (
                t.Param(
                    "DstAddr",
                    t.NWK,
                    "Short address of the device that will process the "
                    "mgmt leave (remote or self)",
                ),
                t.Param(
                    "IEEE",
                    t.EUI64,
                    (
                        "The 64-bit IEEE address of the entity to be removed from the "
                        "network or 0x0000000000000000 if the device removes itself "
                        "from the network."
                    ),
                ),
                t.Param(
                    "RemoveChildren_Rejoin",
                    LeaveOptions,
                    "Specifies actions to be performed by "
                    "device when leaving the network.",
                ),
            )
        ),
        rsp_schema=t.STATUS_SCHEMA,
    )

    # request the Management Direct Join Request of a designated device
    MgmtDirectJoinReq = t.CommandDef(
        t.CommandType.SREQ,
        0x35,
        req_schema=t.Schema(
            (
                t.Param("Dst", t.NWK, "Short address of the device to join"),
                t.Param("IEEE", t.EUI64, "IEEE address of the device to join"),
                t.Param("Capabilities", t.uint8_t, "MAC Capabilities"),
            )
        ),
        rsp_schema=t.STATUS_SCHEMA,
    )

    # set the Permit Join for the destination device
    MgmtPermitJoinReq = t.CommandDef(
        t.CommandType.SREQ,
        0x36,
        req_schema=t.Schema(
            (
                t.Param(
                    "AddrMode", t.AddrMode, "Address mode of DST: short or broadcast"
                ),
                t.Param("Dst", t.NWK, "Short address of the device to join"),
                t.Param(
                    "Duration", t.uint8_t, "Specifies the duration to permit joining"
                ),
                t.Param(
                    "TCSignificance",
                    t.uint8_t,
                    "Trust Center Significance  -- unused in the code!",
                ),
            )
        ),
        rsp_schema=t.STATUS_SCHEMA,
    )

    # allow updating of network configuration parameters or to request information
    # from devices on network conditions in the local operating environment
    MgmtNWKUpdateReq = t.CommandDef(
        t.CommandType.SREQ,
        0x37,
        req_schema=t.Schema(
            (
                t.Param("Dst", t.NWK, "Short address of the destination device"),
                t.Param("DstAddrMode", t.AddrMode, "Destination Address mode"),
                t.Param("Channels", t.Channels, "Bitmask of channels to scan"),
                t.Param("ScanDuration", t.uint8_t, "Scanning time"),
                t.Param(
                    "ScanCount",
                    t.uint8_t,
                    "The number of energy scans to be conducted and reported",
                ),
                t.Param(
                    "NwkManagerAddr",
                    t.NWK,
                    "NWK address for the device with the Network Manager bit set",
                ),
            )
        ),
        rsp_schema=t.STATUS_SCHEMA,
    )

    # register for a ZDO callback
    MsgCallbackRegister = t.CommandDef(
        t.CommandType.SREQ,
        0x3E,
        req_schema=t.Schema(
            (
                t.Param(
                    "ClusterId",
                    t.ClusterId,
                    "Cluster id for which to receive ZDO callback",
                ),
            )
        ),
        rsp_schema=t.STATUS_SCHEMA,
    )

    # de-register for a ZDO callback
    MsgCallbackRemove = t.CommandDef(
        t.CommandType.SREQ,
        0x3F,
        req_schema=t.Schema(
            (
                t.Param(
                    "ClusterId",
                    t.ClusterId,
                    "Cluster id for which to receive ZDO callback",
                ),
            )
        ),
        rsp_schema=t.STATUS_SCHEMA,
    )

    # starts the device in the network
    StartupFromApp = t.CommandDef(
        t.CommandType.SREQ,
        0x40,
        req_schema=t.Schema((t.Param("StartDelay", t.uint16_t, "Startup delay"),)),
        rsp_schema=t.Schema((t.Param("State", StartupState, "State after startup"),)),
    )

    # Extended version of ZDO to indicate to router devices to create
    # a distributed network
    StartupFromAppExt = t.CommandDef(
        t.CommandType.SREQ,
        0x54,
        req_schema=t.Schema(
            (
                t.Param("StartDelay", t.uint16_t, "Startup delay"),
                t.Param(
                    "Mode", t.Bool, "True -- ZR devices to create a distributed network"
                ),
            )
        ),
        rsp_schema=t.Schema((t.Param("State", StartupState, "State after startup"),)),
    )

    # set the application link key for a given device
    SetLinkKey = t.CommandDef(
        t.CommandType.SREQ,
        0x23,
        req_schema=t.Schema(
            (
                t.Param("NWK", t.NWK, "Short address of the device"),
                t.Param("IEEE", t.EUI64, "Extended address of the device"),
                t.Param("LinkKeyData", t.KeyData, "128bit link key"),
            )
        ),
        rsp_schema=t.STATUS_SCHEMA,
    )

    # remove the application link key for a given device
    RemoveLinkKey = t.CommandDef(
        t.CommandType.SREQ,
        0x24,
        req_schema=t.Schema(
            (t.Param("IEEE", t.EUI64, "Extended address of the device"),)
        ),
        rsp_schema=t.STATUS_SCHEMA,
    )

    # get the application link key for a given device
    GetLinkKey = t.CommandDef(
        t.CommandType.SREQ,
        0x25,
        req_schema=t.Schema(
            (t.Param("IEEE", t.EUI64, "Extended address of the device"),)
        ),
        rsp_schema=t.Schema(
            (
                t.Param(
                    "Status", t.ZDOStatus, "Status is either Success (0) or Failure (1)"
                ),
                t.Param("IEEE", t.EUI64, "Extended address of the device"),
                t.Param("LinkKeyData", t.KeyData, "128bit link key"),
            )
        ),
    )

    # initiate a network discovery (active scan)
    NetworkDiscoveryReq = t.CommandDef(
        t.CommandType.SREQ,
        0x26,
        req_schema=t.Schema(
            (
                t.Param("Channels", t.Channels, "Bitmask of channels to scan"),
                t.Param("ScanDuration", t.uint8_t, "Scanning time"),
            )
        ),
        rsp_schema=t.STATUS_SCHEMA,
    )

    # request the device to join itself to a parent device on a network
    JoinReq = t.CommandDef(
        t.CommandType.SREQ,
        0x27,
        req_schema=t.Schema(
            (
                t.Param(
                    "LogicalChannel", t.uint8_t, "Channel where the PAN is located"
                ),
                t.Param("PanId", t.PanId, "The PAN Id to join."),
                t.Param("ExtendedPanId", t.ExtendedPanId, "64-bit extended PAN ID"),
                t.Param(
                    "ChosenParent",
                    t.NWK,
                    "Short address of the parent device chosen to join",
                ),
                t.Param("Depth", t.uint8_t, "Depth of the parent"),
                t.Param(
                    "StackProfile", t.uint8_t, "Stack profile of the network to use"
                ),
            )
        ),
        rsp_schema=t.STATUS_SCHEMA,
    )

    # set rejoin backoff duration and rejoin scan duration for an end device
    SetRejoinParams = t.CommandDef(
        t.CommandType.SREQ,
        # in documentation CmdId=0x26 which conflict with discover req
        0x28,
        req_schema=t.Schema(
            (
                t.Param(
                    "BackoffDuraation",
                    t.uint32_t,
                    "Rejoin backoff  duration for end device",
                ),
                t.Param(
                    "ScanDuration", t.uint32_t, "Rejoin scan duration for end device"
                ),
            )
        ),
        rsp_schema=t.STATUS_SCHEMA,
    )

    # handles the ZDO security add link key extension message
    SecAddLinkKey = t.CommandDef(
        t.CommandType.SREQ,
        0x42,
        req_schema=t.Schema(
            (
                t.Param("NWK", t.NWK, "Short address of the device"),
                t.Param("IEEE", t.EUI64, "Extended address of the device"),
                t.Param("LinkKeyData", t.KeyData, "128bit link key"),
            )
        ),
        rsp_schema=t.STATUS_SCHEMA,
    )

    # handle the ZDO security entry lookup extended extension message
    SecEntryLookupExt = t.CommandDef(
        t.CommandType.SREQ,
        0x43,
        req_schema=t.Schema(
            (
                t.Param("IEEE", t.EUI64, "Extended address of the device"),
                t.Param("Entry", SecurityEntry, "Valid entry"),
            )
        ),
        rsp_schema=t.Schema(
            (
                t.Param("AMI", t.uint16_t, "Address manager index"),
                t.Param("KeyNVID", t.uint16_t, "Index to link key table in NV"),
                t.Param("Option", t.uint8_t, "Authentication option for device"),
            )
        ),
    )

    # handle the ZDO security remove device extended extension message
    SecDeviceRemove = t.CommandDef(
        t.CommandType.SREQ,
        0x44,
        req_schema=t.Schema(
            (t.Param("IEEE", t.EUI64, "Extended address of the device"),)
        ),
        rsp_schema=t.STATUS_SCHEMA,
    )

    # handle the ZDO route discovery extension message
    ExtRouteDisc = t.CommandDef(
        t.CommandType.SREQ,
        0x45,
        req_schema=t.Schema(
            (
                t.Param("Dst", t.NWK, "Short address of the destination"),
                t.Param("Options", RouteDiscoveryOptions, "Route options"),
                t.Param("Radius", t.uint8_t, "Broadcast radius"),
            )
        ),
        rsp_schema=t.STATUS_SCHEMA,
    )

    # handle the ZDO route check extension messags
    ExtRouteChk = t.CommandDef(
        t.CommandType.SREQ,
        0x46,
        req_schema=t.Schema(
            (
                t.Param("Dst", t.NWK, "Short address of the destination"),
                t.Param("RtStatus", t.uint8_t, "Status value for routing entries"),
                t.Param("Options", t.uint8_t, "Route options"),
            )
        ),
        rsp_schema=t.STATUS_SCHEMA,
    )

    # handle the ZDO extended remove group extension message
    ExtRemoveGroup = t.CommandDef(
        t.CommandType.SREQ,
        0x47,
        req_schema=t.Schema(
            (
                t.Param("Endpoint", t.uint8_t, "Endpoint to look for"),
                t.Param("GroupId", t.GroupId, "ID to look for group"),
            )
        ),
        rsp_schema=t.STATUS_SCHEMA,
    )

    # handle the ZDO extended remove all group extension message
    ExtRemoveAllGroups = t.CommandDef(
        t.CommandType.SREQ,
        0x48,
        req_schema=t.Schema((t.Param("Endpoint", t.uint8_t, "Endpoint to look for"),)),
        rsp_schema=t.STATUS_SCHEMA,
    )

    # handle the ZDO extension find all groups for endpoint message
    ExtFindAllGroupsEndpoint = t.CommandDef(
        t.CommandType.SREQ,
        0x49,
        req_schema=t.Schema(
            (
                t.Param("Endpoint", t.uint8_t, "Endpoint to look for"),
                # this parameter does not make sense
                t.Param("Groups", t.uint16_t, "List to hold group IDs"),
            )
        ),
        rsp_schema=t.Schema((t.Param("Groups", GroupIdList, "List of Group IDs"),)),
    )

    # handle the ZDO extension find group message
    ExtFindGroup = t.CommandDef(
        t.CommandType.SREQ,
        0x4A,
        req_schema=t.Schema(
            (
                t.Param("Endpoint", t.uint8_t, "Endpoint to look for"),
                t.Param("GroupId", t.GroupId, "ID to look for group"),
            )
        ),
        rsp_schema=t.Schema((t.Param("Group", t.Bytes, "Group information"),)),
    )

    # handle the ZDO extension add group message
    ExtAddGroup = t.CommandDef(
        t.CommandType.SREQ,
        0x4B,
        req_schema=t.Schema(
            (
                t.Param("Endpoint", t.uint8_t, "Endpoint to look for"),
                t.Param("GroupId", t.GroupId, "ID to look for group"),
                t.Param("GroupName", t.CharacterString, "Group name"),
            )
        ),
        rsp_schema=t.STATUS_SCHEMA,
    )

    # handle the ZDO extension count all groups message
    ExtCountAllGroups = t.CommandDef(
        t.CommandType.SREQ,
        0x4C,
        req_schema=t.Schema(),
        rsp_schema=t.Schema(
            (t.Param("GroupCount", t.uint8_t, "Total number of groups"),)
        ),
    )

    # handle the ZDO extension Get/Set RxOnIdle to ZMac message
    ExtRxIdle = t.CommandDef(
        t.CommandType.SREQ,
        0x4D,
        req_schema=t.Schema(
            (
                t.Param("SetFlag", t.uint8_t, "Set or get value"),
                t.Param("SetValue", t.uint8_t, "Value to be set to ZMac message"),
            )
        ),
        rsp_schema=t.STATUS_SCHEMA,
    )

    # handle the ZDO security update network key extension message
    ExtUpdateNwkKey = t.CommandDef(
        t.CommandType.SREQ,
        0x4E,
        req_schema=t.Schema(
            (
                t.Param("Dst", t.NWK, "Short address of the destination"),
                t.Param("KeySeqNum", t.uint8_t, "Key sequence number"),
                t.Param("Key", t.KeyData, "Network key"),
            )
        ),
        rsp_schema=t.STATUS_SCHEMA,
    )

    # handle the ZDO security switch network key extension message
    ExtSwitchNwkKey = t.CommandDef(
        t.CommandType.SREQ,
        0x4F,
        req_schema=t.Schema(
            (
                t.Param("Dst", t.NWK, "Short address of the destination"),
                t.Param("KeySeqNum", t.uint8_t, "Key sequence number"),
            )
        ),
        rsp_schema=t.STATUS_SCHEMA,
    )

    # handle the ZDO extension network message
    ExtNwkInfo = t.CommandDef(
        t.CommandType.SREQ,
        0x50,
        req_schema=t.Schema(),
        rsp_schema=t.Schema(
            (
                t.Param("Dst", t.NWK, "Short address of the destination"),
                t.Param("PanId", t.PanId, "The PAN Id to join."),
                t.Param("ParentNWK", t.NWK, "Short address of the parent"),
                t.Param("ExtendedPanId", t.ExtendedPanId, "64-bit extended PAN ID"),
                t.Param("ParentIEEE", t.EUI64, "IEEE address of the parent"),
                t.Param("Channel", t.Channels, "Current Channel"),
            )
        ),
    )

    # handle the ZDO extension Security Manager APS Remove Request message
    ExtSecApsRemoveReq = t.CommandDef(
        t.CommandType.SREQ,
        0x51,
        req_schema=t.Schema(
            (
                t.Param("NWK", t.NWK, "Short address of the device"),
                t.Param("IEEE", t.EUI64, "IEEE address of the device"),
                t.Param("ParentNWK", t.NWK, "Short address of the parent"),
            )
        ),
        rsp_schema=t.STATUS_SCHEMA,
    )

    # forces a network concentrator change by resetting zgConcentratorEnable and
    # zgConcentratorDiscoveryTime from NV and set nwk event
    ForceConcentratorChange = t.CommandDef(
        t.CommandType.SREQ, 0x52, req_schema=t.Schema(), rsp_schema=t.Schema()
    )

    # set parameters not settable through NV
    ExtSetParams = t.CommandDef(
        t.CommandType.SREQ,
        0x53,
        req_schema=t.Schema(
            (t.Param("UseMulticast", t.Bool, "Set or reset of multicast"),)
        ),
        rsp_schema=t.STATUS_SCHEMA,
    )

    # handle ZDO network address of interest request
    NwkAddrOfInterestReq = t.CommandDef(
        t.CommandType.SREQ,
        0x29,
        req_schema=t.Schema(
            (
                t.Param("NWK", t.NWK, "Short address of the destination"),
                t.Param(
                    "NWKAddrOfInterest",
                    t.NWK,
                    "Short address of the device being queried",
                ),
                t.Param(
                    "Cmd",
                    t.uint8_t,
                    "A valid Cluser ID command as specified by profile",
                ),
            )
        ),
        rsp_schema=t.STATUS_SCHEMA,
    )

    # ZDO Callbacks
    # return the results to the NwkAddrReq
    NwkAddrRsp = t.CommandDef(
        t.CommandType.AREQ,
        0x80,
        req_schema=t.Schema(),
        rsp_schema=t.Schema(
            (
                t.Param(
                    "Status", t.ZDOStatus, "Status is either Success (0) or Failure (1)"
                ),
                t.Param("IEEE", t.EUI64, "Extended address of the source device"),
                t.Param("NWK", t.NWK, "Short address of the source device"),
                t.Param(
                    "Index",
                    t.uint8_t,
                    "Starting index into the list of associated devices",
                ),
                t.Param("Devices", t.NWKList, "List of the associated devices"),
            )
        ),
    )

    # return the results to the IEEEAddrReq
    IEEEAddrRsp = t.CommandDef(
        t.CommandType.AREQ,
        0x81,
        rsp_schema=t.Schema(
            (
                t.Param(
                    "Status", t.ZDOStatus, "Status is either Success (0) or Failure (1)"
                ),
                t.Param("IEEE", t.EUI64, "Extended address of the source device"),
                t.Param("NWK", t.NWK, "Short address of the source device"),
                t.Param(
                    "Index",
                    t.uint8_t,
                    "Starting index into the list of associated devices",
                ),
                t.Param("Devices", t.NWKList, "List of the associated devices"),
            )
        ),
    )

    # return the results to the NodeDescReq
    NodeDescRsp = t.CommandDef(
        t.CommandType.AREQ,
        0x82,
        rsp_schema=t.Schema(
            (
                t.Param("Src", t.NWK, "The message’s source network address."),
                t.Param(
                    "Status",
                    t.ZDOStatus,
                    "This field indicates either SUCCESS or FAILURE.",
                ),
                t.Param("NWK", t.NWK, "Device’s short address of this Node descriptor"),
                t.Param("NodeDescriptor", NullableNodeDescriptor, "Node descriptor"),
            )
        ),
    )

    # return the results to the PowerDescReq
    PowerDescRsp = t.CommandDef(
        t.CommandType.AREQ,
        0x83,
        rsp_schema=t.Schema(
            (
                t.Param("Src", t.NWK, "message's source network address"),
                t.Param(
                    "Status", t.ZDOStatus, "Status is either Success (0) or Failure (1)"
                ),
                t.Param("NWK", t.NWK, "Short address of the device response describes"),
                t.Param(
                    "PowerDescriptor",
                    zigpy.zdo.types.PowerDescriptor,
                    "Power descriptor response",
                ),
            )
        ),
    )

    # return the results to the SimpleDescReq
    SimpleDescRsp = t.CommandDef(
        t.CommandType.AREQ,
        0x84,
        rsp_schema=t.Schema(
            (
                t.Param("Src", t.NWK, "message's source network address"),
                t.Param(
                    "Status", t.ZDOStatus, "Status is either Success (0) or Failure (1)"
                ),
                t.Param("NWK", t.NWK, "Short address of the device response describes"),
                t.Param(
                    "SimpleDescriptor",
                    zigpy.zdo.types.SizePrefixedSimpleDescriptor,
                    "Simple descriptor",
                ),
            )
        ),
    )

    # return the results to the ActiveEpReq
    ActiveEpRsp = t.CommandDef(
        t.CommandType.AREQ,
        0x85,
        rsp_schema=t.Schema(
            (
                t.Param("Src", t.NWK, "message's source network address"),
                t.Param(
                    "Status", t.ZDOStatus, "Status is either Success (0) or Failure (1)"
                ),
                t.Param("NWK", t.NWK, "Short address of the device response describes"),
                t.Param("ActiveEndpoints", EndpointList, "Active endpoints list"),
            )
        ),
    )

    # return the results to the MatchDescReq
    MatchDescRsp = t.CommandDef(
        t.CommandType.AREQ,
        0x86,
        rsp_schema=t.Schema(
            (
                t.Param("Src", t.NWK, "message's source network address"),
                t.Param(
                    "Status", t.ZDOStatus, "Status is either Success (0) or Failure (1)"
                ),
                t.Param("NWK", t.NWK, "Short address of the device response describes"),
                t.Param("MatchList", EndpointList, "Endpoints list"),
            )
        ),
    )

    # return the results to the ComplexDescReq
    ComplexDescRsp = t.CommandDef(
        t.CommandType.AREQ,
        0x87,
        rsp_schema=t.Schema(
            (
                t.Param("Src", t.NWK, "message's source network address"),
                t.Param(
                    "Status", t.ZDOStatus, "Status is either Success (0) or Failure (1)"
                ),
                t.Param("NWK", t.NWK, "Short address of the device response describes"),
                t.Param("ComplexDesc", t.ShortBytes, "Complex descriptor"),
            )
        ),
    )

    # return the results to the UserDescReq
    UserDescRsp = t.CommandDef(
        t.CommandType.AREQ,
        0x88,
        rsp_schema=t.Schema(
            (
                t.Param("Src", t.NWK, "message's source network address"),
                t.Param(
                    "Status", t.ZDOStatus, "Status is either Success (0) or Failure (1)"
                ),
                t.Param("NWK", t.NWK, "Short address of the device response describes"),
                t.Param("UserDesc", t.ShortBytes, "User descriptor"),
            )
        ),
    )

    # notify the user when the device receives a user descriptor
    UserDescCnf = t.CommandDef(
        t.CommandType.AREQ,
        0x89,
        rsp_schema=t.Schema(
            (
                t.Param("Src", t.NWK, "message's source network address"),
                t.Param(
                    "Status", t.ZDOStatus, "Status is either Success (0) or Failure (1)"
                ),
                t.Param("NWK", t.NWK, "Short address of the device response describes"),
            )
        ),
    )

    # return the results to the ServerDiscReq
    ServerDiscRsp = t.CommandDef(
        t.CommandType.AREQ,
        0x8A,
        rsp_schema=t.Schema(
            (
                t.Param("Src", t.NWK, "message's source network address"),
                t.Param(
                    "Status", t.ZDOStatus, "Status is either Success (0) or Failure (1)"
                ),
                t.Param("ServerMask", t.ZDOStatus, "Server mask response"),
            )
        ),
    )

    # return the results to the EndDeviceBindReq
    EndDeviceBindRsp = t.CommandDef(
        t.CommandType.AREQ,
        0xA0,
        rsp_schema=t.Schema(
            (
                t.Param("Src", t.NWK, "message's source network address"),
                t.Param(
                    "Status", t.ZDOStatus, "Status is either Success (0) or Failure (1)"
                ),
            )
        ),
    )

    # return the results to the BindReq
    BindRsp = t.CommandDef(
        t.CommandType.AREQ,
        0xA1,
        rsp_schema=t.Schema(
            (
                t.Param("Src", t.NWK, "message's source network address"),
                t.Param(
                    "Status", t.ZDOStatus, "Status is either Success (0) or Failure (1)"
                ),
            )
        ),
    )

    # return the results to the UnBindReq
    UnBindRsp = t.CommandDef(
        t.CommandType.AREQ,
        0xA2,
        rsp_schema=t.Schema(
            (
                t.Param("Src", t.NWK, "message's source network address"),
                t.Param(
                    "Status", t.ZDOStatus, "Status is either Success (0) or Failure (1)"
                ),
            )
        ),
    )

    # return the results to the MgmtNwkDiscReq
    MgmtNwkDiscRsp = t.CommandDef(
        t.CommandType.AREQ,
        0xB0,
        rsp_schema=t.Schema(
            (
                t.Param("Src", t.NWK, "message's source network address"),
                t.Param(
                    "Status", t.ZDOStatus, "Status is either Success (0) or Failure (1)"
                ),
                t.Param("NetworkCount", t.uint8_t, "Total number of entries available"),
                t.Param("Index", t.uint8_t, "Where the response starts"),
                t.Param("Networks", NetworkList, "Discovered networks list"),
            )
        ),
    )

    # return the results to the MgmtLqiReq
    MgmtLqiRsp = t.CommandDef(
        t.CommandType.AREQ,
        0xB1,
        rsp_schema=t.Schema(
            (
                t.Param("Src", t.NWK, "message's source network address"),
                t.Param(
                    "Status", t.ZDOStatus, "Status is either Success (0) or Failure (1)"
                ),
                t.Param("Neighbours", zigpy.zdo.types.Neighbors, "Neighbours"),
            )
        ),
    )

    # return the results to the MgmtRtgReq
    MgmtRtgRsp = t.CommandDef(
        t.CommandType.AREQ,
        0xB2,
        rsp_schema=t.Schema(
            (
                t.Param("Src", t.NWK, "message's source network address"),
                t.Param(
                    "Status", t.ZDOStatus, "Status is either Success (0) or Failure (1)"
                ),
                t.Param("Routes", zigpy.zdo.types.Routes, "Routes"),
            )
        ),
    )

    # return the results to the MgmtBingReq
    MgmtBindRsp = t.CommandDef(
        t.CommandType.AREQ,
        0xB3,
        rsp_schema=t.Schema(
            (
                t.Param("Src", t.NWK, "message's source network address"),
                t.Param(
                    "Status", t.ZDOStatus, "Status is either Success (0) or Failure (1)"
                ),
                t.Param(
                    "BindTableEntries",
                    t.uint8_t,
                    "Total number of entries available on the device",
                ),
                t.Param("Index", t.uint8_t, "Index where the response starts"),
                t.Param("BindTable", BindEntryList, "list of BindEntries"),
            )
        ),
    )

    # return the results to the MgmtLeaveReq
    MgmtLeaveRsp = t.CommandDef(
        t.CommandType.AREQ,
        0xB4,
        rsp_schema=t.Schema(
            (
                t.Param("Src", t.NWK, "message's source network address"),
                t.Param(
                    "Status", t.ZDOStatus, "Status is either Success (0) or Failure (1)"
                ),
            )
        ),
    )

    # return the results to the MgmtDirectJoinReq
    MgmtDirectJoinRsp = t.CommandDef(
        t.CommandType.AREQ,
        0xB5,
        rsp_schema=t.Schema(
            (
                t.Param("Src", t.NWK, "message's source network address"),
                t.Param(
                    "Status", t.ZDOStatus, "Status is either Success (0) or Failure (1)"
                ),
            )
        ),
    )

    # return the results to the MgmtPermitJoinReq
    MgmtPermitJoinRsp = t.CommandDef(
        t.CommandType.AREQ,
        0xB6,
        rsp_schema=t.Schema(
            (
                t.Param("Src", t.NWK, "message's source network address"),
                t.Param(
                    "Status", t.ZDOStatus, "Status is either Success (0) or Failure (1)"
                ),
            )
        ),
    )

    # indicates ZDO state change
    StateChangeInd = t.CommandDef(
        t.CommandType.AREQ,
        0xC0,
        rsp_schema=t.Schema((t.Param("State", t.DeviceState, "New ZDO state"),)),
    )

    # indicates the ZDO End Device Announce
    EndDeviceAnnceInd = t.CommandDef(
        t.CommandType.AREQ,
        0xC1,
        rsp_schema=t.Schema(
            (
                t.Param("Src", t.NWK, "Source address of the message."),
                t.Param("NWK", t.NWK, "Specifies the device’s short address"),
                t.Param(
                    "IEEE",
                    t.EUI64,
                    "Extended address of the device generating the request",
                ),
                t.Param("Capabilities", MACCapabilities, "MAC Capabilities"),
            )
        ),
    )

    # indicates that Match Descriptor Response has been sent
    MatchDescRspSent = t.CommandDef(
        t.CommandType.AREQ,
        0xC2,
        rsp_schema=t.Schema(
            (
                t.Param("NWK", t.NWK, "Device's network address"),
                t.Param("InputClusters", t.ClusterIdList, "Input cluster id list"),
                t.Param("OutputClusters", t.ClusterIdList, "Output cluster id list"),
            )
        ),
    )

    # default message for error status
    StatusErrorRsp = t.CommandDef(
        t.CommandType.AREQ,
        0xC3,
        rsp_schema=t.Schema(
            (
                t.Param("Src", t.NWK, "message's source network address"),
                t.Param(
                    "Status", t.ZDOStatus, "Status is either Success (0) or Failure (1)"
                ),
            )
        ),
    )

    # indication to inform host device the receipt of a source route to a given device
    SrcRtgInd = t.CommandDef(
        t.CommandType.AREQ,
        0xC4,
        rsp_schema=t.Schema(
            (
                t.Param(
                    "DstAddr",
                    t.NWK,
                    "Network address of the destination of the source route",
                ),
                t.Param("Relays", t.NWKList, "List of relay devices"),
            )
        ),
    )

    # indication to inform host device the receipt of a beacon notification
    BeaconNotifyInd = t.CommandDef(
        t.CommandType.AREQ,
        0xC5,
        rsp_schema=t.Schema((t.Param("Beacons", BeaconList, "Beacons list"),)),
    )

    # inform the host device of a ZDO join request result
    JoinCnf = t.CommandDef(
        t.CommandType.AREQ,
        0xC6,
        rsp_schema=t.Schema(
            (
                t.Param(
                    "Status", t.ZDOStatus, "Status is either Success (0) or Failure (1)"
                ),
                t.Param("Nwk", t.NWK, "device's network address"),
                t.Param("ParentNwk", t.NWK, "Parent's network address"),
            )
        ),
    )

    # indication to inform host device the completion of network discovery scan
    NwkDiscoveryCnf = t.CommandDef(t.CommandType.AREQ, 0xC7, rsp_schema=t.STATUS_SCHEMA)

    # an indication to inform the host of a device leaving the network
    LeaveInd = t.CommandDef(
        t.CommandType.AREQ,
        0xC9,
        rsp_schema=t.Schema(
            (
                t.Param(
                    "NWK", t.NWK, "Short address of the source of the leave indication"
                ),
                t.Param(
                    "IEEE",
                    t.EUI64,
                    "IEEE address of the source of the leave indication",
                ),
                t.Param("Request", t.Bool, "True -- request, False -- indication"),
                t.Param("Remove", t.Bool, "True -- Remove children"),
                t.Param("Rejoin", t.Bool, "True -- Rejoin"),
            )
        ),
    )

    # ZDO callback for a Cluster Id that the host requested to receive
    # with a MsgCallbackRegister request
    MsgCbIncoming = t.CommandDef(
        t.CommandType.AREQ,
        0xFF,
        rsp_schema=t.Schema(
            (
                t.Param("Src", t.NWK, "Source address of the ZDO message"),
                t.Param(
                    "IsBroadcast",
                    t.Bool,
                    "Indicates whether the message was a broadcast",
                ),
                t.Param("ClusterId", t.ClusterId, "Cluster Id of this ZDO message"),
                t.Param("SecurityUse", t.uint8_t, "Not used"),
                t.Param("TSN", t.uint8_t, "Transaction sequence number"),
                t.Param(
                    "MacDst", t.NWK, "Mac destination short address of the ZDO message"
                ),
                t.Param(
                    "Data",
                    t.Bytes,
                    "Data that corresponds to the cluster ID of the message",
                ),
            )
        ),
    )

    # a ZDO callback for TC Device Indication
    TCDevInd = t.CommandDef(
        t.CommandType.AREQ,
        0xCA,
        rsp_schema=t.Schema(
            (
                t.Param("SrcNwk", t.NWK, "device's network address"),
                t.Param("SrcIEEE", t.EUI64, "IEEE address of the source"),
                t.Param("ParentNwk", t.NWK, "Parent's network address"),
            )
        ),
    )

    # a ZDO callback for Permit Join Indication
    PermitJoinInd = t.CommandDef(
        t.CommandType.AREQ,
        0xCB,
        rsp_schema=t.Schema((t.Param("Duration", t.uint8_t, "Permit join duration"),)),
    )



import enum
import typing


class Bytes(bytes):
    def serialize(self) -> "Bytes":
        return self

    @classmethod
    def deserialize(cls, data: bytes) -> typing.Tuple["Bytes", bytes]:
        return cls(data), b""

    def __repr__(self) -> str:
        # Reading byte sequences like \x200\x21 is extremely annoying
        # compared to \x20\x30\x21
        escaped = "".join(f"\\x{b:02X}" for b in self)

        return f"b'{escaped}'"

    __str__ = __repr__


def serialize_list(objects) -> Bytes:
    return Bytes(b"".join([o.serialize() for o in objects]))


class int_t(int):
    _signed = True
    _size = None

    def __new__(cls, value):
        instance = int.__new__(cls, value)

        if instance._signed is not None and instance._size is not None:
            # It's a concrete int_t type, check to make sure it's valid
            instance.serialize()

        return instance

    def serialize(self) -> bytes:
        try:
            return self.to_bytes(self._size, "little", signed=self._signed)
        except OverflowError as e:
            # OverflowError is not a subclass of ValueError, making it annoying to catch
            raise ValueError(str(e)) from e

    @classmethod
    def deserialize(cls, data: bytes) -> typing.Tuple["int_t", bytes]:
        if len(data) < cls._size:
            raise ValueError(f"Data is too short to contain {cls._size} bytes")

        r = cls.from_bytes(data[: cls._size], "little", signed=cls._signed)
        data = data[cls._size :]
        return r, data


class int8s(int_t):
    _size = 1


class int16s(int_t):
    _size = 2


class int24s(int_t):
    _size = 3


class int32s(int_t):
    _size = 4


class int40s(int_t):
    _size = 5


class int48s(int_t):
    _size = 6


class int56s(int_t):
    _size = 7


class int64s(int_t):
    _size = 8


class uint_t(int_t):
    _signed = False


class uint8_t(uint_t):
    _size = 1


class uint16_t(uint_t):
    _size = 2


class uint24_t(uint_t):
    _size = 3


class uint32_t(uint_t):
    _size = 4


class uint40_t(uint_t):
    _size = 5


class uint48_t(uint_t):
    _size = 6


class uint56_t(uint_t):
    _size = 7


class uint64_t(uint_t):
    _size = 8


class ShortBytes(Bytes):
    _header = uint8_t

    def serialize(self) -> "Bytes":
        return self._header(len(self)).serialize() + self

    @classmethod
    def deserialize(cls, data: bytes) -> typing.Tuple[Bytes, bytes]:
        length, data = cls._header.deserialize(data)
        if length > len(data):
            raise ValueError(f"Data is too short to contain {length} bytes of data")
        return cls(data[:length]), data[length:]


class LongBytes(ShortBytes):
    _header = uint16_t


class TypedListMeta(list):
    _item_type = None
    _length = None

    def serialize(self) -> bytes:
        if self._length is not None and len(self) != self._length:
            raise ValueError(
                f"Invalid length for {self!r}: expected {self._length}, got {len(self)}"
            )

        return b"".join([self._item_type(i).serialize() for i in self])

    @classmethod
    def deserialize(cls, data: bytes):
        raise NotImplementedError()  # pragma: no cover


class LVList(TypedListMeta):
    def __init_subclass__(cls, *, item_type, length_type) -> None:
        super().__init_subclass__()
        cls._item_type = item_type
        cls._header = length_type

    def serialize(self) -> bytes:
        assert self._item_type is not None
        return self._header(len(self)).serialize() + super().serialize()

    @classmethod
    def deserialize(cls, data: bytes):
        assert cls._item_type is not None
        length, data = cls._header.deserialize(data)
        r = cls()
        for i in range(length):
            item, data = cls._item_type.deserialize(data)
            r.append(item)
        return r, data


class FixedList(TypedListMeta):
    def __init_subclass__(cls, *, item_type, length) -> None:
        super().__init_subclass__()
        cls._item_type = item_type
        cls._length = length

    @classmethod
    def deserialize(cls, data):
        assert cls._item_type is not None
        r = cls()
        for i in range(cls._length):
            item, data = cls._item_type.deserialize(data)
            r.append(item)
        return r, data


class HexRepr:
    def __str__(self):
        return ("0x{:0" + str(self._size * 2) + "X}").format(self)

    __repr__ = __str__


class EnumIntFlagMixin:
    """
    Enum does not allow multiple base classes. We turn enum.IntFlag into a mixin, since
    it really doesn't depend on the base class specifically being `int`.
    """

    # Rebind classmethods to our own class
    _missing_ = classmethod(enum.IntFlag._missing_.__func__)
    _create_pseudo_member_ = classmethod(enum.IntFlag._create_pseudo_member_.__func__)

    __or__ = enum.IntFlag.__or__
    __and__ = enum.IntFlag.__and__
    __xor__ = enum.IntFlag.__xor__
    __ror__ = enum.IntFlag.__ror__
    __rand__ = enum.IntFlag.__rand__
    __rxor__ = enum.IntFlag.__rxor__
    __invert__ = enum.IntFlag.__invert__


class enum_uint8(uint8_t, enum.Enum):
    pass


class enum_uint16(uint16_t, enum.Enum):
    pass


class enum_uint24(uint24_t, enum.Enum):
    pass


class enum_uint32(uint32_t, enum.Enum):
    pass


class enum_uint40(uint40_t, enum.Enum):
    pass


class enum_uint48(uint48_t, enum.Enum):
    pass


class enum_uint56(uint56_t, enum.Enum):
    pass


class enum_uint64(uint64_t, enum.Enum):
    pass


class enum_flag_uint8(EnumIntFlagMixin, uint8_t, enum.Flag):
    pass


class enum_flag_uint16(EnumIntFlagMixin, uint16_t, enum.Flag):
    pass


class enum_flag_uint24(EnumIntFlagMixin, uint24_t, enum.Flag):
    pass


class enum_flag_uint32(EnumIntFlagMixin, uint32_t, enum.Flag):
    pass


class enum_flag_uint40(EnumIntFlagMixin, uint40_t, enum.Flag):
    pass


class enum_flag_uint48(EnumIntFlagMixin, uint48_t, enum.Flag):
    pass


class enum_flag_uint56(EnumIntFlagMixin, uint56_t, enum.Flag):
    pass


class enum_flag_uint64(EnumIntFlagMixin, uint64_t, enum.Flag):
    pass



import attr
import typing
import asyncio
import logging
import async_timeout

from collections import defaultdict

import zigpy_znp.commands
import zigpy_znp.types as t
import zigpy_znp.config as conf
import zigpy_znp.commands as c
from zigpy_znp.types import nvids

from zigpy_znp import uart
from zigpy_znp.frames import GeneralFrame
from zigpy_znp.exceptions import CommandNotRecognized, InvalidCommandResponse


LOGGER = logging.getLogger(__name__)


def _deduplicate_commands(commands):
    # Command matching as a relation forms a partially ordered set.
    # To avoid triggering our callbacks multiple times per packet, we
    # should remove redundant partial commands.
    maximal_commands = []

    for command in commands:
        for index, other_command in enumerate(maximal_commands):
            if other_command.matches(command):
                # If the other command matches us, we are redundant
                break
            elif command.matches(other_command):
                # If we match another command, we replace it
                maximal_commands[index] = command
                break
            else:
                # Otherwise, we keep looking
                pass  # pragma: no cover
        else:
            # If we matched nothing and nothing matched us, we extend the list
            maximal_commands.append(command)

    # The start of each chain is the maximal element
    return tuple(maximal_commands)


@attr.s(frozen=True)
class BaseResponseListener:
    matching_commands: typing.Tuple[t.CommandBase] = attr.ib(
        converter=_deduplicate_commands
    )

    @matching_commands.validator
    def check(self, attribute, commands):
        if not commands:
            raise ValueError("Listener must have at least one command")

    def matching_headers(self):
        return {response.header for response in self.matching_commands}

    def resolve(self, response: t.CommandBase) -> bool:
        if not any(c.matches(response) for c in self.matching_commands):
            return False

        if not self._resolve(response):
            return False

        return True

    def _resolve(self, response: t.CommandBase) -> bool:
        """
        Implemented by subclasses to handle matched commands.

        Return value indicates whether or not the listener has actually resolved,
        which can sometimes be unavoidable.
        """
        raise NotImplementedError()  # pragma: no cover

    def cancel(self):
        """
        Implement by subclasses to cancel the listener.

        Return value indicates whether or not the listener is cancelable.
        """
        raise NotImplementedError()  # pragma: no cover


@attr.s(frozen=True)
class OneShotResponseListener(BaseResponseListener):
    future: asyncio.Future = attr.ib(
        default=attr.Factory(lambda: asyncio.get_running_loop().create_future())
    )

    def _resolve(self, response: t.CommandBase) -> bool:
        if self.future.done():
            # This happens if the UART receives multiple packets during the same
            # event loop step and all of them match this listener. Our Future's
            # add_done_callback will not fire synchronously and thus the listener
            # is never properly removed. This isn't going to break anything.
            LOGGER.debug("Future already has a result set: %s", self.future)
            return False

        self.future.set_result(response)
        return True

    def cancel(self):
        if not self.future.done():
            self.future.cancel()

        return True


@attr.s(frozen=True)
class CallbackResponseListener(BaseResponseListener):
    callback: typing.Callable[[t.CommandBase], typing.Any] = attr.ib()

    def _resolve(self, response: t.CommandBase) -> bool:
        try:
            result = self.callback(response)

            # Run coroutines in the background
            if asyncio.iscoroutine(result):
                asyncio.create_task(result)
        except Exception:
            LOGGER.warning(
                "Caught an exception while executing callback", exc_info=True
            )

        # Returning False could cause our callback to be called multiple times in a row
        return True

    def cancel(self):
        # You can't cancel a callback
        return False


class ZNP:
    def __init__(self, config: conf.ConfigType):
        self._uart = None
        self._app = None
        self._config = config

        self._response_listeners = defaultdict(list)
        self._sync_request_lock = asyncio.Lock()

    def set_application(self, app):
        assert self._app is None
        self._app = app

    @property
    def _port_path(self) -> str:
        return self._config[conf.CONF_DEVICE][conf.CONF_DEVICE_PATH]

    async def connect(self) -> None:
        assert self._uart is None

        self._uart = await uart.connect(self._config[conf.CONF_DEVICE], self)
        LOGGER.debug("Testing connection to %s", self._uart.transport.serial.name)

        try:
            # Make sure that our port works
            await self.request(c.Sys.Ping.Req())
        except Exception:
            self._uart = None
            raise

        LOGGER.debug(
            "Connected to %s at %s baud",
            self._uart.transport.serial.name,
            self._uart.transport.serial.baudrate,
        )

    def _cancel_all_listeners(self) -> None:
        for header, listeners in self._response_listeners.items():
            for listener in listeners:
                listener.cancel()

    def connection_lost(self, exc) -> None:
        LOGGER.debug("We were disconnected from %s: %s", self._port_path, exc)

        self._uart = None
        self.close()

        if self._app is not None:
            self._app.connection_lost(exc)

    def close(self) -> None:
        if self._uart is not None:
            self._uart.close()
            self._uart = None

        self._cancel_all_listeners()

    def _remove_listener(self, listener: BaseResponseListener) -> None:
        LOGGER.trace("Removing listener %s", listener)

        for header in listener.matching_headers():
            self._response_listeners[header].remove(listener)

            if not self._response_listeners[header]:
                LOGGER.trace("Cleaning up empty listener list for header %s", header)
                del self._response_listeners[header]

        total_listeners = sum(map(len, self._response_listeners.values()))
        LOGGER.trace("There are %d listeners remaining", total_listeners)

    def frame_received(self, frame: GeneralFrame) -> None:
        """
        Called when a frame has been received.
        Can be called multiple times in a single event loop step.
        """

        command_cls = zigpy_znp.commands.COMMANDS_BY_ID[frame.header]
        command = command_cls.from_frame(frame)

        LOGGER.debug("Received command: %s", command)

        matched = False

        for listener in self._response_listeners[command.header]:
            if not listener.resolve(command):
                LOGGER.trace("%s does not match %s", command, listener)
                continue

            matched = True
            LOGGER.trace("%s matches %s", command, listener)

        if not matched:
            LOGGER.warning("Received an unhandled command: %s", command)

    def callback_for_responses(self, responses, callback) -> None:
        listener = CallbackResponseListener(responses, callback=callback)

        LOGGER.trace("Creating callback %s", listener)

        for header in listener.matching_headers():
            self._response_listeners[header].append(listener)

    def callback_for_response(self, response, callback) -> None:
        return self.callback_for_responses([response], callback)

    def wait_for_responses(self, responses) -> asyncio.Future:
        listener = OneShotResponseListener(responses)

        LOGGER.trace("Creating one-shot listener %s", listener)

        for header in listener.matching_headers():
            self._response_listeners[header].append(listener)

        # Remove the listener when the future is done, not only when it gets a result
        listener.future.add_done_callback(lambda _: self._remove_listener(listener))

        return listener.future

    def wait_for_response(self, response: t.CommandBase) -> asyncio.Future:
        return self.wait_for_responses([response])

    async def request(self, request, **response_params):
        if type(request) is not request.Req:
            raise ValueError(f"Cannot send a command that isn't a request: {request!r}")

        if request.Rsp:
            renamed_response_params = {}

            for param, value in response_params.items():
                if not param.startswith("Rsp"):
                    raise KeyError(
                        f"All response params must start with 'Rsp': {param!r}"
                    )

                renamed_response_params[param.replace("Rsp", "", 1)] = value

            # Construct our response before we send the request so that we fail early
            partial_response = request.Rsp(partial=True, **renamed_response_params)
        elif response_params:
            raise ValueError(
                f"Command has no response so response_params={response_params} "
                f"will have no effect"
            )

        LOGGER.debug("Sending request: %s", request)

        # If our request has no response, we cannot wait for one
        if not request.Rsp:
            LOGGER.debug("Request has no response, not waiting for one.")
            self._uart.send(request.to_frame())
            return

        # We should only be sending one SREQ at a time, according to the spec
        async with self._sync_request_lock:
            # We need to create the response listener before we send the request
            response_future = self.wait_for_responses(
                [
                    request.Rsp(partial=True),
                    c.RPCError.CommandNotRecognized.Rsp(
                        partial=True, RequestHeader=request.header
                    ),
                ]
            )
            self._uart.send(request.to_frame())

            # We should get a SRSP in a reasonable amount of time
            async with async_timeout.timeout(
                self._config[conf.CONF_ZNP_CONFIG][conf.CONF_SREQ_TIMEOUT]
            ):
                # We lock until either a sync response is seen or an error occurs
                response = await response_future

        if isinstance(response, c.RPCError.CommandNotRecognized.Rsp):
            raise CommandNotRecognized(f"Fatal request error: {response}")

        # If the sync response we got is not what we wanted, this is an error
        if not partial_response.matches(response):
            raise InvalidCommandResponse(
                f"Expected SRSP response {partial_response}, got {response}", response
            )

        return response

    async def request_callback_rsp(self, *, request, callback, **response_params):
        """
        Sends a [SA]REQ request and waits for its AREQ response. A bug-free version of:

            req_rsp = await req
            callback_rsp = await req_callback
        """

        callback_response = self.wait_for_response(callback)
        response = self.request(request, **response_params)

        await response
        return await callback_response

    async def nvram_write(
        self, nv_id: nvids.BaseNvIds, value, *, offset: t.uint8_t = 0
    ):
        # While unpythonic, explicit type checking here means we can detect overflows
        if not isinstance(nv_id, nvids.BaseNvIds):
            raise ValueError(
                "The nv_id param must be an instance of BaseNvIds. "
                "Extend one of the tables in zigpy_znp.types.nvids."
            )

        if not isinstance(value, bytes):
            value = value.serialize()

        return await self.request(
            c.Sys.OSALNVWrite.Req(Id=nv_id, Offset=offset, Value=t.ShortBytes(value)),
            RspStatus=t.Status.SUCCESS,
        )

    async def nvram_read(
        self, nv_id: nvids.BaseNvIds, *, offset: t.uint8_t = 0
    ) -> bytes:
        response = await self.request(
            c.Sys.OSALNVRead.Req(Id=nv_id, Offset=offset), RspStatus=t.Status.SUCCESS,
        )

        return response.Value



"""This interface allows tester to interact with the simple API interface."""

import zigpy_znp.types as t


class SAPI(t.CommandsBase, subsystem=t.Subsystem.SAPI):
    # reset the device by using a soft reset (i.e. a jump to the reset vector) vice a
    # hardware reset (i.e. watchdog reset.)
    ZBSystemReset = t.CommandDef(t.CommandType.AREQ, 0x09, req_schema=t.Schema())

    # start the ZigBee stack
    ZBStartReq = t.CommandDef(
        t.CommandType.SREQ, 0x00, req_schema=t.Schema(), rsp_schema=t.Schema()
    )

    # control the joining permissions and thus allows or disallows new devices
    # from joining the network
    ZBPermitJoiningReq = t.CommandDef(
        t.CommandType.SREQ,
        0x08,
        req_schema=t.Schema(
            (
                t.Param(
                    "NWK",
                    t.NWK,
                    (
                        "Short address of the device for which the joining "
                        "permissions should be set"
                    ),
                ),
                t.Param(
                    "duration",
                    t.uint8_t,
                    "amount of time in seconds to allow new device to join",
                ),
            )
        ),
        rsp_schema=t.STATUS_SCHEMA,
    )

    # establishes or removes a ‘binding’ between two devices. Once bound, an
    # application can send messages to a device by referencing the commandId
    # for the binding
    ZBBindDevice = t.CommandDef(
        t.CommandType.SREQ,
        0x01,
        req_schema=t.Schema(
            (
                t.Param("Create", t.Bool, "True to create binding, False to remove"),
                t.Param("CommandId", t.uint16_t, "The identifier of the binding"),
                t.Param("IEEE", t.EUI64, "IEEE address of the device to bind to"),
            )
        ),
        rsp_schema=t.Schema(),
    )

    # puts the device into the Allow Binding Mode for a given period of time. A peer
    # device can establish a binding to a device in the Allow Binding Mode by calling
    # zb_BindDevice with a destination address of NULL
    ZBAllowBind = t.CommandDef(
        t.CommandType.SREQ,
        0x02,
        req_schema=t.Schema(
            (
                t.Param(
                    "duration",
                    t.uint8_t,
                    "amount of time in seconds to allow new device to join",
                ),
            )
        ),
        rsp_schema=t.Schema(),
    )

    # initiates transmission of data to a peer device
    ZBSendDataRequest = t.CommandDef(
        t.CommandType.SREQ,
        0x03,
        req_schema=t.Schema(
            (
                t.Param("Destination", t.NWK, "Short address of the destination"),
                t.Param(
                    "CommandId", t.uint16_t, "The command id to send with the message"
                ),
                t.Param(
                    "Handle",
                    t.uint8_t,
                    "A handle used to Identify the send data request",
                ),
                t.Param("Ack", t.Bool, "True if requesting ACK from the destination"),
                t.Param(
                    "Radius", t.uint8_t, "The max number of hops the packet can travel"
                ),
                t.Param("Data", t.ShortBytes, "Data"),
            )
        ),
        rsp_schema=t.Schema(),
    )

    # get a configuration property from nonvolatile memory
    ZBReadConfiguration = t.CommandDef(
        t.CommandType.SREQ,
        0x04,
        req_schema=t.Schema(
            (
                t.Param(
                    "ConfigId", t.uint8_t, "ID of the configuration property to read"
                ),
            )
        ),
        rsp_schema=t.Schema(
            (
                t.Param(
                    "Status", t.Status, "Status is either Success (0) or Failure (1)"
                ),
                t.Param(
                    "ConfigId", t.uint8_t, "ID of the configuration property to read"
                ),
                t.Param("Value", t.ShortBytes, "Value"),
            )
        ),
    )

    # write a configuration property from nonvolatile memory
    ZBWriteConfiguration = t.CommandDef(
        t.CommandType.SREQ,
        0x05,
        req_schema=t.Schema(
            (
                t.Param(
                    "Status", t.Status, "Status is either Success (0) or Failure (1)"
                ),
                t.Param(
                    "ConfigId", t.uint8_t, "ID of the configuration property to read"
                ),
                t.Param("Value", t.ShortBytes, "Value"),
            )
        ),
        rsp_schema=t.STATUS_SCHEMA,
    )

    # retrieves a Device Information Property
    ZBGetDeviceInfo = t.CommandDef(
        t.CommandType.SREQ,
        0x06,
        req_schema=t.Schema(
            (t.Param("Param", t.uint8_t, "The identifier for deice information"),)
        ),
        rsp_schema=t.Schema(
            (
                t.Param("Param", t.uint8_t, "The identifier for deice information"),
                t.Param("Value", t.uint16_t, "Value"),
            )
        ),
    )

    #  determine the short address for a device in the network. The device
    # initiating a call to zb_FindDeviceRequest and the device being discovered must
    # both be a member of the same network. When the search is complete,
    # the zv_FindDeviceConfirm callback function is called
    ZBFindDeviceReq = t.CommandDef(
        t.CommandType.SREQ,
        0x07,
        req_schema=t.Schema(
            (t.Param("SearchKey", t.EUI64, "IEEE address of the device"),)
        ),
        rsp_schema=t.Schema(),
    )

    # SAPI CallBacks
    # this callback is called by the ZigBee stack after a start request
    # operation completes
    ZBStartConfirm = t.CommandDef(t.CommandType.AREQ, 0x80, req_schema=t.STATUS_SCHEMA)

    # This callback is called by the ZigBee stack after a bind operation completes
    ZBBindConfirm = t.CommandDef(
        t.CommandType.AREQ,
        0x81,
        rsp_schema=t.Schema(
            (
                t.Param(
                    "CommandId", t.uint16_t, "The command id to send with the message"
                ),
                t.Param(
                    "Status", t.Status, "Status is either Success (0) or Failure (1)"
                ),
            )
        ),
    )

    # This callback indicates another device attempted to bind to this device
    ZBAllowBindConfirm = t.CommandDef(
        t.CommandType.AREQ,
        0x82,
        rsp_schema=t.Schema(
            (t.Param("Source", t.NWK, "Source Nwk of the device attempted to bind"),)
        ),
    )

    # This callback indicates the data has been sent
    ZBSendConfirm = t.CommandDef(
        t.CommandType.AREQ,
        0x83,
        rsp_schema=t.Schema(
            (
                t.Param(
                    "Handle",
                    t.uint8_t,
                    "A handle used to Identify the send data request",
                ),
                t.Param(
                    "Status", t.Status, "Status is either Success (0) or Failure (1)"
                ),
            )
        ),
    )

    # This callback is called asynchronously by the ZigBee stack to notify the
    # application when data is received from a peer device
    ZBRecieveDataInd = t.CommandDef(
        t.CommandType.AREQ,
        0x87,
        rsp_schema=t.Schema(
            (
                t.Param("Source", t.NWK, "NWK address of the source device"),
                t.Param(
                    "CommandId", t.uint16_t, "The command id associated with the data"
                ),
                t.Param("Data", t.LongBytes, "Data"),
            )
        ),
    )

    # This callback is called by the ZigBee stack when a find device operation
    # completes
    ZBFindDeviceConfirm = t.CommandDef(
        t.CommandType.AREQ,
        0x85,
        rsp_schema=t.Schema(
            (
                t.Param(
                    "SearchType", t.uint8_t, "The type of search that was performed"
                ),
                t.Param(
                    "SearchKey", t.uint16_t, "Value that the search was executed on"
                ),
                t.Param("Result", t.EUI64, "The result of the search"),
            )
        ),
    )



import zigpy_znp.types as t


class App(t.CommandsBase, subsystem=t.Subsystem.APP):
    # This command is sent to the target in order to test the functions defined
    # for individual applications.
    # This command sends a raw data to an application
    Msg = t.CommandDef(
        t.CommandType.SREQ,
        0x00,
        req_schema=t.Schema(
            (
                t.Param(
                    "Endpoint",
                    t.uint8_t,
                    "Application endpoint of the outgoing message",
                ),
                t.Param(
                    "DstAddr", t.NWK, "Destination address of the outgoing message"
                ),
                t.Param(
                    "DstEndpoint",
                    t.uint8_t,
                    "Destination endpoint of the outgoing message",
                ),
                t.Param("ClusterId", t.ClusterId, "Cluster ID"),
                t.Param("Data", t.ShortBytes, "Data request"),
            )
        ),
        rsp_schema=t.STATUS_SCHEMA,
    )

    # This command is used by tester to issue user’s defined commands to the
    # application
    UserTest = t.CommandDef(
        t.CommandType.SREQ,
        0x01,
        req_schema=t.Schema(
            (
                t.Param(
                    "SrcEndpoint",
                    t.uint8_t,
                    "Source Endpoint of the user-defined command",
                ),
                t.Param(
                    "CommandId", t.uint16_t, "Command Id of the user-defined command"
                ),
                t.Param("Parameter1", t.uint16_t, "Parameter #1 of the command"),
                t.Param("Parameter2", t.uint16_t, "Parameter #2 of the command"),
            )
        ),
        rsp_schema=t.STATUS_SCHEMA,
    )



import attr
import typing
import functools

import zigpy_znp.types as t
from zigpy_znp.exceptions import InvalidFrame


@attr.s
class GeneralFrame:
    header: t.CommandHeader = attr.ib(converter=t.CommandHeader)
    data: t.Bytes = attr.ib(factory=t.Bytes, converter=t.Bytes)

    @property
    def length(self) -> t.uint8_t:
        """Length of the frame."""
        return t.uint8_t(len(self.data))

    @classmethod
    def deserialize(cls, data):
        """Deserialize frame and sanity checks."""
        length, data = t.uint8_t.deserialize(data)

        if length > 250:
            raise InvalidFrame(f"Frame length cannot exceed 250 bytes. Got: {length}")

        if len(data) < length + 2:
            raise InvalidFrame(f"Data is too short for {cls.__name__}")

        header, data = t.CommandHeader.deserialize(data)
        payload, data = data[:length], data[length:]
        return cls(header, payload), data

    @data.validator
    def data_validator(self, attribute, value):
        """Len of data should not exceed 250 bytes."""
        if len(value) > 250:
            raise ValueError(f"data length: {len(value)} exceeds max 250")

    def serialize(self) -> bytes:
        """Serialize Frame."""
        return self.length.serialize() + self.header.serialize() + self.data.serialize()


@attr.s
class TransportFrame:
    """Transport frame."""

    SOF = t.uint8_t(0xFE)

    payload: GeneralFrame = attr.ib()

    @classmethod
    def deserialize(cls, data: bytes) -> typing.Tuple["TransportFrame", bytes]:
        """Deserialize frame."""
        sof, data = t.uint8_t.deserialize(data)

        if sof != cls.SOF:
            raise InvalidFrame(
                f"Expected frame to start with SOF 0x{cls.SOF:02X}, got 0x{sof:02X}"
            )

        gen_frame, data = GeneralFrame.deserialize(data)
        checksum, data = t.uint8_t.deserialize(data)

        frame = cls(gen_frame)

        if frame.checksum() != checksum:
            raise InvalidFrame(
                f"Invalid frame checksum for data {gen_frame}: "
                f"expected 0x{frame.checksum():02X}, got 0x{checksum:02X}"
            )

        return frame, data

    def checksum(self) -> t.uint8_t:
        """Calculate FCS on the payload."""
        checksum = functools.reduce(lambda a, b: a ^ b, self.payload.serialize())
        return t.uint8_t(checksum)

    def serialize(self) -> bytes:
        """Serialize data."""
        return (
            self.SOF.serialize()
            + self.payload.serialize()
            + self.checksum().serialize()
        )



"""Green Power devices and Green Power infrastructure interface."""

import zigpy_znp.types as t


class ZGP(t.CommandsBase, subsystem=t.Subsystem.ZGP):
    DataReq = t.CommandDef(
        t.CommandType.SREQ,
        0x01,
        req_schema=t.Schema(
            (
                t.Param(
                    "Action", t.Bool, "True/False -- add/remove GPDF into the queue"
                ),
                t.Param("TXOptions", t.uint8_t, "ZGP TX Options"),
                t.Param(
                    "ApplicationId",
                    t.uint8_t,
                    (
                        "ApplicationID of the GPD to which the ASDU will be sent; "
                        "ApplicationID 0x00 indicates the usage of the SrcID; "
                        "ApplicationID 0x02 indicates the usage of the GPD IEEE"
                    ),
                ),
                t.Param(
                    "SrcId",
                    t.uint32_t,
                    "The identifier of the GPD entity to which the ASDU will be sent",
                ),
                t.Param("IEEE", t.EUI64, "IEEE address"),
                t.Param(
                    "Endpoint",
                    t.uint8_t,
                    "GPD ep used in combination with GPD IEEE for APPid of 0b010",
                ),
                t.Param("CommandId", t.uint8_t, "GPD command id"),
                t.Param("ASDU", t.ShortBytes, "GPD ASDU"),
                t.Param(
                    "Handle", t.uint8_t, "GPEP handle to match req to confirmation"
                ),
                t.Param("LifeTime", t.uint24_t, "The lifetime of the packet"),
            )
        ),
        rsp_schema=t.STATUS_SCHEMA,
    )

    # send security data into # the dGP stub
    SecRsp = t.CommandDef(
        t.CommandType.SREQ,
        0x02,
        req_schema=t.Schema(
            (
                t.Param(
                    "Status",
                    t.uint8_t,
                    "The status code as returned by the GP endpoint",
                ),
                t.Param(
                    "Handle", t.uint8_t, "GPEP handle to match req to confirmation"
                ),
                t.Param(
                    "ApplicationId",
                    t.uint8_t,
                    (
                        "ApplicationID of the GPD to which the ASDU will be sent; "
                        "ApplicationID 0x00 indicates the usage of the SrcID; "
                        "ApplicationID 0x02 indicates the usage of the GPD IEEE"
                    ),
                ),
                t.Param(
                    "SrcId",
                    t.uint32_t,
                    "The identifier of the GPD entity to which the ASDU will be sent",
                ),
                t.Param("IEEE", t.EUI64, "IEEE address"),
                t.Param(
                    "Endpoint",
                    t.uint8_t,
                    "GPD ep used in combination with GPD IEEE for APPid of 0b010",
                ),
                t.Param(
                    "SecurityLevel", t.uint8_t, "Security level for GPDF processing"
                ),
                t.Param("KeyType", t.uint8_t, "The security key type"),
                t.Param("KeyData", t.KeyData, "Security key"),
                t.Param("FrameCounter", t.uint32_t, "The security frame counter value"),
            )
        ),
        rsp_schema=t.STATUS_SCHEMA,
    )

    # ZGP Callbacks
    # Green power confirm is a message that provides a mechanism for the Green Power
    # EndPoint in the host processor to understand the status of a previous request
    # to send a GPDF
    DataCnf = t.CommandDef(
        t.CommandType.AREQ,
        0x05,
        req_schema=t.Schema(
            (
                t.Param(
                    "Status",
                    t.uint8_t,
                    "The status code as returned by the GP endpoint",
                ),
                t.Param("Handle", t.uint8_t, "handle to match req to confirmation"),
            )
        ),
    )

    # This message provides a mechanism for dGP stub to request security data from the
    # Green Power EndPoint in the host processor
    SecReq = t.CommandDef(
        t.CommandType.AREQ,
        0x03,
        req_schema=t.Schema(
            (
                t.Param(
                    "ApplicationId",
                    t.uint8_t,
                    (
                        "ApplicationID of the GPD to which the ASDU will be sent; "
                        "ApplicationID 0x00 indicates the usage of the SrcID; "
                        "ApplicationID 0x02 indicates the usage of the GPD IEEE"
                    ),
                ),
                t.Param(
                    "SrcId",
                    t.uint32_t,
                    "The identifier of the GPD entity to which the ASDU will be sent",
                ),
                t.Param("IEEE", t.EUI64, "IEEE address"),
                t.Param(
                    "Endpoint",
                    t.uint8_t,
                    "GPD ep used in combination with GPD IEEE for APPid of 0b010",
                ),
                t.Param(
                    "SecurityLevel", t.uint8_t, "Security level for GPDF processing"
                ),
                t.Param("KeyType", t.uint8_t, "The security key type"),
                t.Param("FrameCounter", t.uint32_t, "The security frame counter value"),
                t.Param(
                    "Handle", t.uint8_t, "dGP stub handle to match req to confirmation"
                ),
            )
        ),
    )

    # This message provides a mechanism for identifying and conveying a received
    # GPDF to the Green Power EndPoint in the host processor
    DataInd = t.CommandDef(
        t.CommandType.AREQ,
        0x04,
        req_schema=t.Schema(
            (
                t.Param(
                    "Status", t.uint8_t, "The status code as returned by the dGP stub"
                ),
                t.Param(
                    "RSSI",
                    t.int8s,
                    "The RSSI delivered by the MAC on receipt of this frame",
                ),
                t.Param("LQI", t.uint8_t, "Link quality measured during reception"),
                t.Param(
                    "SeqNum", t.uint8_t, "The sequence number from MAC header of MPDU"
                ),
                t.Param(
                    "SrcAddrMode", t.AddrMode, "The source addressing mode of MPDU"
                ),
                t.Param("SrcPanId", t.PanId, "The source PAN Id"),
                # ToDo: ugh, this could be ieee depending on addressing mode
                t.Param("SrcNWK", t.NWK, "The source address of the GPD entity"),
                t.Param(
                    "DstAddrMode", t.AddrMode, "The destination addressing mode of MPDU"
                ),
                t.Param("DstPanId", t.PanId, "The destination PAN Id"),
                # ToDo: ugh, this could be ieee depending on addressing mode
                t.Param("DstNWK", t.NWK, "The destination address of the GPD entity"),
                t.Param("MPDU", t.ShortBytes, "GP MPDU"),
            )
        ),
    )



import typing
import numbers

import voluptuous as vol
from zigpy.config import (  # noqa: F401
    CONFIG_SCHEMA,
    SCHEMA_DEVICE,
    CONF_DEVICE,
    CONF_DEVICE_PATH,
    CONF_NWK,
    CONF_NWK_CHANNEL,
    CONF_NWK_CHANNELS,
    CONF_NWK_EXTENDED_PAN_ID,
    CONF_NWK_PAN_ID,
    CONF_NWK_KEY,
    CONF_NWK_TC_ADDRESS,
    CONF_NWK_TC_LINK_KEY,
    CONF_NWK_UPDATE_ID,
    cv_boolean,
)

ConfigType = typing.Dict[str, typing.Any]

VolPositiveNumber = vol.All(numbers.Real, vol.Range(min=0))

CONF_DEVICE_BAUDRATE = "baudrate"
CONF_DEVICE_FLOW_CONTROL = "flow_control"

SCHEMA_DEVICE = SCHEMA_DEVICE.extend(
    {
        vol.Optional(CONF_DEVICE_BAUDRATE, default=115_200): int,
        vol.Optional(CONF_DEVICE_FLOW_CONTROL, default=None): vol.In(
            ("hardware", "software", None)
        ),
    }
)


CONF_ZNP_CONFIG = "znp_config"
CONF_TX_POWER = "tx_power"
CONF_SREQ_TIMEOUT = "sync_request_timeout"
CONF_AUTO_RECONNECT_RETRY_DELAY = "auto_reconnect_retry_delay"

CONFIG_SCHEMA = CONFIG_SCHEMA.extend(
    {
        vol.Required(CONF_DEVICE): SCHEMA_DEVICE,
        vol.Optional(CONF_ZNP_CONFIG, default={}): vol.Schema(
            {
                vol.Optional(CONF_TX_POWER, default=None): vol.Any(
                    None, vol.All(int, vol.Range(min=-22, max=19))
                ),
                vol.Optional(CONF_SREQ_TIMEOUT, default=5): VolPositiveNumber,
                vol.Optional(
                    CONF_AUTO_RECONNECT_RETRY_DELAY, default=5
                ): VolPositiveNumber,
            }
        ),
    }
)



import attr
import typing


class Struct:
    """Structure based on attr."""

    @classmethod
    def deserialize(cls, data: bytes):
        """Deserialize structure."""

        args = []

        for field_type in cls.schema():
            arg, data = field_type.deserialize(data)
            args.append(arg)

        return cls(*args), data

    @classmethod
    def schema(cls):
        """Return schema of the class."""
        return (a.type for a in attr.fields(cls))

    def serialize(self):
        """Serialize Structure."""
        values = self.as_tuple()
        return b"".join(v.serialize() for v in values)

    def as_tuple(self):
        """Return tuple of fields values."""
        names = (f.name for f in attr.fields(self.__class__))
        return (getattr(self, n) for n in names)

    @staticmethod
    def converter(_type) -> typing.Callable:
        """Pass through converter."""

        def converter(input):
            if isinstance(input, _type):
                return input
            try:
                return _type(*input)
            except TypeError:
                return _type(input)

        return converter



import os
import typing
import asyncio
import logging
import itertools
import async_timeout

import zigpy.util
import zigpy.types
import zigpy.config
import zigpy.application
import zigpy.profiles
import zigpy.zcl.foundation

from zigpy.zdo.types import ZDOCmd, ZDOHeader, CLUSTERS as ZDO_CLUSTERS

from zigpy.types import ExtendedPanId, deserialize as list_deserialize
from zigpy.zcl.clusters.security import IasZone
from zigpy.exceptions import DeliveryError

import zigpy_znp.config as conf
import zigpy_znp.types as t
import zigpy_znp.commands as c
from zigpy_znp.exceptions import InvalidCommandResponse

from zigpy_znp.api import ZNP
from zigpy_znp.types.nvids import NwkNvIds


ZDO_ENDPOINT = 0
ZDO_REQUEST_TIMEOUT = 10  # seconds
DATA_CONFIRM_TIMEOUT = 5  # seconds
LOGGER = logging.getLogger(__name__)


ZDO_COMPLEX_CONVERTERS = {
    ZDOCmd.Node_Desc_req: (
        (
            lambda addr, device, NWKAddrOfInterest: c.ZDO.NodeDescReq.Req(
                DstAddr=addr, NWKAddrOfInterest=NWKAddrOfInterest
            )
        ),
        (lambda addr: c.ZDO.NodeDescRsp.Callback(partial=True, Src=addr)),
        (
            lambda rsp: (
                ZDOCmd.Node_Desc_rsp,
                {
                    "Status": rsp.Status,
                    "NWKAddrOfInterest": rsp.NWK,
                    "NodeDescriptor": rsp.NodeDescriptor,
                },
            )
        ),
    ),
    ZDOCmd.Active_EP_req: (
        (
            lambda addr, device, NWKAddrOfInterest: c.ZDO.ActiveEpReq.Req(
                DstAddr=addr, NWKAddrOfInterest=NWKAddrOfInterest
            )
        ),
        (lambda addr: c.ZDO.ActiveEpRsp.Callback(partial=True, Src=addr)),
        (
            lambda rsp: (
                ZDOCmd.Active_EP_rsp,
                {
                    "Status": rsp.Status,
                    "NWKAddrOfInterest": rsp.NWK,
                    "ActiveEPList": rsp.ActiveEndpoints,
                },
            )
        ),
    ),
    ZDOCmd.Simple_Desc_req: (
        (
            lambda addr, device, NWKAddrOfInterest, EndPoint: (
                c.ZDO.SimpleDescReq.Req(
                    DstAddr=addr, NWKAddrOfInterest=NWKAddrOfInterest, Endpoint=EndPoint
                )
            )
        ),
        (lambda addr: c.ZDO.SimpleDescRsp.Callback(partial=True, Src=addr)),
        (
            lambda rsp: (
                ZDOCmd.Simple_Desc_rsp,
                {
                    "Status": rsp.Status,
                    "NWKAddrOfInterest": rsp.NWK,
                    "SimpleDescriptor": rsp.SimpleDescriptor,
                },
            )
        ),
    ),
    ZDOCmd.Mgmt_Leave_req: (
        (
            lambda addr, device, DeviceAddress, Options: c.ZDO.MgmtLeaveReq.Req(
                DstAddr=addr,
                IEEE=device.ieee,
                RemoveChildren_Rejoin=c.zdo.LeaveOptions(Options),
            )
        ),
        (lambda addr: c.ZDO.MgmtLeaveRsp.Callback(partial=True, Src=addr)),
        (lambda rsp: (ZDOCmd.Mgmt_Leave_rsp, {"Status": rsp.Status})),
    ),
    ZDOCmd.Bind_req: (
        (
            lambda addr, device, SrcAddress, SrcEndpoint, ClusterID, DstAddress: (
                c.ZDO.BindReq.Req(
                    Dst=addr,
                    Src=SrcAddress,
                    SrcEndpoint=SrcEndpoint,
                    ClusterId=ClusterID,
                    Address=DstAddress,
                )
            )
        ),
        (lambda addr: c.ZDO.BindRsp.Callback(partial=True, Src=addr)),
        (lambda rsp: (ZDOCmd.Bind_rsp, {"Status": rsp.Status})),
    ),
}


class ControllerApplication(zigpy.application.ControllerApplication):
    SCHEMA = conf.CONFIG_SCHEMA
    SCHEMA_DEVICE = conf.SCHEMA_DEVICE

    def __init__(self, config: conf.ConfigType):
        super().__init__(config=conf.CONFIG_SCHEMA(config))

        self._znp = None

        # It's easier to deal with this if it's never None
        self._reconnect_task = asyncio.Future()
        self._reconnect_task.cancel()

    @classmethod
    async def probe(cls, device_config: conf.ConfigType) -> bool:
        znp = ZNP(conf.CONFIG_SCHEMA({conf.CONF_DEVICE: device_config}))
        LOGGER.debug("Probing %s", znp._port_path)

        try:
            await znp.connect()
            return True
        except Exception as e:
            LOGGER.warning(
                "Failed to probe ZNP radio with config %s", device_config, exc_info=e
            )
            return False
        finally:
            znp.close()

    def _receive_zdo_message(
        self, cluster: ZDOCmd, *, tsn: t.uint8_t, sender, **zdo_kwargs
    ) -> None:
        field_names, field_types = ZDO_CLUSTERS[cluster]
        assert set(zdo_kwargs) == set(field_names)

        zdo_args = []

        for f_name, f_type in zip(field_names, field_types):
            # XXX: bugfix for zigpy optional struct bug
            if f_type.__name__ == "Optional":
                f_type = f_type.__mro__[1]

            zdo_args.append(f_type(zdo_kwargs[f_name]))

        message = t.serialize_list([t.uint8_t(tsn)] + zdo_args)

        LOGGER.trace("Pretending we received a ZDO message: %s", message)
        self.handle_message(
            sender=sender,
            profile=zigpy.profiles.zha.PROFILE_ID,
            cluster=cluster,
            src_ep=ZDO_ENDPOINT,
            dst_ep=ZDO_ENDPOINT,
            message=message,
        )

    def on_zdo_relays_message(self, msg: c.ZDO.SrcRtgInd.Callback) -> None:
        LOGGER.info("ZDO device relays: %s", msg)
        device = self.get_device(nwk=msg.DstAddr)
        device.relays = msg.Relays

    def on_zdo_device_announce(self, msg: c.ZDO.EndDeviceAnnceInd.Callback) -> None:
        LOGGER.info("ZDO device announce: %s", msg)

        self._receive_zdo_message(
            cluster=ZDOCmd.Device_annce,
            tsn=0xFF,
            sender=self.get_device(ieee=msg.IEEE),
            NWKAddr=msg.NWK,
            IEEEAddr=msg.IEEE,
            Capability=msg.Capabilities,
        )

    def on_tc_device_join(self, msg: c.ZDO.TCDevInd.Callback) -> None:
        LOGGER.info("TC device join: %s", msg)
        self.handle_join(nwk=msg.SrcNwk, ieee=msg.SrcIEEE, parent_nwk=msg.ParentNwk)

    def on_zdo_device_leave(self, msg: c.ZDO.LeaveInd.Callback) -> None:
        LOGGER.info("ZDO device left: %s", msg)
        self.handle_leave(nwk=msg.NWK, ieee=msg.IEEE)

    def on_af_message(self, msg: c.AF.IncomingMsg.Callback) -> None:
        try:
            device = self.get_device(nwk=msg.SrcAddr)
        except KeyError:
            LOGGER.warning(
                "Received an AF message from an unknown device: 0x%04x", msg.SrcAddr
            )
            return

        device.radio_details(lqi=msg.LQI, rssi=None)

        self.handle_message(
            sender=device,
            profile=zigpy.profiles.zha.PROFILE_ID,
            cluster=msg.ClusterId,
            src_ep=msg.SrcEndpoint,
            dst_ep=msg.DstEndpoint,
            message=msg.Data,
        )

    async def shutdown(self):
        """Shutdown application."""

        self._reconnect_task.cancel()
        self._znp.close()

    def _bind_callbacks(self, api):
        api.callback_for_response(
            c.AF.IncomingMsg.Callback(partial=True), self.on_af_message
        )

        # ZDO requests need to be handled explicitly
        api.callback_for_response(
            c.ZDO.EndDeviceAnnceInd.Callback(partial=True), self.on_zdo_device_announce,
        )

        api.callback_for_response(
            c.ZDO.TCDevInd.Callback.Callback(partial=True), self.on_tc_device_join,
        )

        api.callback_for_response(
            c.ZDO.LeaveInd.Callback(partial=True), self.on_zdo_device_leave
        )

        api.callback_for_response(
            c.ZDO.SrcRtgInd.Callback(partial=True), self.on_zdo_relays_message
        )

    async def _reconnect(self) -> None:
        for attempt in itertools.count(start=1):
            LOGGER.debug(
                "Trying to reconnect to %s, attempt %d",
                self._config[conf.CONF_DEVICE][conf.CONF_DEVICE_PATH],
                attempt,
            )

            try:
                await self.startup()

                return
            except Exception as e:
                LOGGER.error("Failed to reconnect", exc_info=e)
                await asyncio.sleep(
                    self._config[conf.CONF_ZNP_CONFIG][
                        conf.CONF_AUTO_RECONNECT_RETRY_DELAY
                    ]
                )

    def connection_lost(self, exc):
        self._znp = None

        # exc=None means that the connection was closed
        if exc is None:
            LOGGER.debug("Connection was purposefully closed. Not reconnecting.")
            return

        # Reconnect in the background using our previously-detected port.
        LOGGER.debug("Starting background reconnection task")

        self._reconnect_task.cancel()
        self._reconnect_task = asyncio.create_task(self._reconnect())

    async def _register_endpoint(
        self,
        endpoint,
        profile_id=zigpy.profiles.zha.PROFILE_ID,
        device_id=zigpy.profiles.zha.DeviceType.CONFIGURATION_TOOL,
        device_version=0x00,
        latency_req=c.af.LatencyReq.NoLatencyReqs,
        input_clusters=[],
        output_clusters=[],
    ):
        return await self._znp.request(
            c.AF.Register.Req(
                Endpoint=endpoint,
                ProfileId=profile_id,
                DeviceId=device_id,
                DeviceVersion=device_version,
                LatencyReq=latency_req,
                InputClusters=input_clusters,
                OutputClusters=output_clusters,
            ),
            RspStatus=t.Status.SUCCESS,
        )

    async def startup(self, auto_form=False):
        """Perform a complete application startup"""

        znp = ZNP(self.config)
        znp.set_application(self)
        self._bind_callbacks(znp)
        await znp.connect()

        self._znp = znp

        # XXX: To make sure we don't switch to the wrong device upon reconnect,
        #      update our config to point to the last-detected port.
        if self._config[conf.CONF_DEVICE][conf.CONF_DEVICE_PATH] == "auto":
            self._config[conf.CONF_DEVICE][
                conf.CONF_DEVICE_PATH
            ] = self._znp._uart.transport.serial.name

        # It's better to configure these explicitly than rely on the NVRAM defaults
        await self._znp.nvram_write(NwkNvIds.CONCENTRATOR_ENABLE, t.Bool(True))
        await self._znp.nvram_write(NwkNvIds.CONCENTRATOR_DISCOVERY, t.uint8_t(120))
        await self._znp.nvram_write(NwkNvIds.CONCENTRATOR_RC, t.Bool(True))
        await self._znp.nvram_write(NwkNvIds.SRC_RTG_EXPIRY_TIME, t.uint8_t(255))
        await self._znp.nvram_write(NwkNvIds.NWK_CHILD_AGE_ENABLE, t.Bool(False))

        # Reset to make the above NVRAM writes take effect.
        # This also ensures any previously-started network joins don't continue.
        await self._reset()

        try:
            is_configured = (
                await self._znp.nvram_read(NwkNvIds.HAS_CONFIGURED_ZSTACK3)
            ) == b"\x55"
        except InvalidCommandResponse as e:
            assert e.response.Status == t.Status.INVALID_PARAMETER
            is_configured = False

        if not is_configured and not auto_form:
            raise RuntimeError("Cannot start application, network is not formed")
        elif auto_form and is_configured:
            LOGGER.info("ZNP is already configured, no need to form a network.")
        elif auto_form and not is_configured:
            await self.form_network()

        if self.config[conf.CONF_ZNP_CONFIG][conf.CONF_TX_POWER] is not None:
            dbm = self.config[conf.CONF_ZNP_CONFIG][conf.CONF_TX_POWER]

            await self._znp.request(
                c.Sys.SetTxPower.Req(TXPower=dbm), RspStatus=t.Status.SUCCESS
            )

        device_info = await self._znp.request(
            c.Util.GetDeviceInfo.Req(), RspStatus=t.Status.SUCCESS
        )

        self._ieee = device_info.IEEE

        if device_info.DeviceState != t.DeviceState.StartedAsCoordinator:
            # Start the application and wait until it's ready
            await self._znp.request_callback_rsp(
                request=c.ZDO.StartupFromApp.Req(StartDelay=100),
                RspState=c.zdo.StartupState.RestoredNetworkState,
                callback=c.ZDO.StateChangeInd.Callback(
                    State=t.DeviceState.StartedAsCoordinator
                ),
            )

        # Get our active endpoints
        endpoints = await self._znp.request_callback_rsp(
            request=c.ZDO.ActiveEpReq.Req(DstAddr=0x0000, NWKAddrOfInterest=0x0000),
            RspStatus=t.Status.SUCCESS,
            callback=c.ZDO.ActiveEpRsp.Callback(partial=True),
        )

        # Clear out the list of active endpoints
        for endpoint in endpoints.ActiveEndpoints:
            await self._znp.request(
                c.AF.Delete.Req(Endpoint=endpoint), RspStatus=t.Status.SUCCESS
            )

        # Register our endpoints
        await self._register_endpoint(endpoint=1)
        await self._register_endpoint(
            endpoint=8,
            device_id=zigpy.profiles.zha.DeviceType.IAS_CONTROL,
            output_clusters=[IasZone.cluster_id],
        )
        await self._register_endpoint(endpoint=11)
        await self._register_endpoint(endpoint=12)
        await self._register_endpoint(
            endpoint=100, profile_id=zigpy.profiles.zll.PROFILE_ID, device_id=0x0005
        )

    async def update_network(
        self,
        *,
        channel: typing.Optional[t.uint8_t] = None,
        channels: typing.Optional[t.Channels] = None,
        extended_pan_id: typing.Optional[t.ExtendedPanId] = None,
        network_key: typing.Optional[t.KeyData] = None,
        pan_id: typing.Optional[t.PanId] = None,
        tc_address: typing.Optional[t.EUI64] = None,
        tc_link_key: typing.Optional[t.KeyData] = None,
        update_id: int = 0,
        reset: bool = True,
    ):
        if (
            channel is not None
            and channels is not None
            and not t.Channels.from_channel_list([channel]) & channels
        ):
            raise ValueError("Channel does not overlap with channel mask")

        if channel is not None:
            LOGGER.warning("Cannot set a specific channel in config: %d", channel)

        if tc_link_key is not None:
            LOGGER.warning("Trust center link key in config is not yet supported")

        if tc_address is not None:
            LOGGER.warning("Trust center address in config is not yet supported")

        if channels is not None:
            await self._znp.request(
                c.Util.SetChannels.Req(Channels=channels), RspStatus=t.Status.SUCCESS,
            )
            await self._znp.request(
                c.AppConfig.BDBSetChannel.Req(IsPrimary=True, Channel=channels),
                RspStatus=t.Status.SUCCESS,
            )
            await self._znp.request(
                c.AppConfig.BDBSetChannel.Req(
                    IsPrimary=False, Channel=t.Channels.NO_CHANNELS
                ),
                RspStatus=t.Status.SUCCESS,
            )

            self._channels = channels

        if pan_id is not None:
            await self._znp.request(
                c.Util.SetPanId.Req(PanId=pan_id), RspStatus=t.Status.SUCCESS
            )

            self._pan_id = pan_id

        if extended_pan_id is not None:
            # There is no Util request to do this
            await self._znp.nvram_write(NwkNvIds.EXTENDED_PAN_ID, extended_pan_id)

            self._ext_pan_id = extended_pan_id

        if network_key is not None:
            await self._znp.request(
                c.Util.SetPreConfigKey.Req(PreConfigKey=network_key),
                RspStatus=t.Status.SUCCESS,
            )

            # XXX: The Util request does not actually write to this NV address
            await self._znp.nvram_write(NwkNvIds.PRECFGKEYS_ENABLE, t.Bool(True))

        if reset:
            # We have to reset afterwards
            await self._reset()

    async def _reset(self):
        await self._znp.request_callback_rsp(
            request=c.Sys.ResetReq.Req(Type=t.ResetType.Soft),
            callback=c.Sys.ResetInd.Callback(partial=True),
        )

    async def form_network(self):
        # These options are read only on startup so we perform a soft reset right after
        await self._znp.nvram_write(
            NwkNvIds.STARTUP_OPTION, t.StartupOptions.ClearState
        )

        # XXX: the undocumented `znpBasicCfg` request can do this
        await self._znp.nvram_write(
            NwkNvIds.LOGICAL_TYPE, t.DeviceLogicalType.Coordinator
        )

        pan_id = self.config[conf.CONF_NWK][conf.CONF_NWK_PAN_ID]
        extended_pan_id = self.config[conf.CONF_NWK][conf.CONF_NWK_EXTENDED_PAN_ID]

        await self.update_network(
            channels=self.config[conf.CONF_NWK][conf.CONF_NWK_CHANNELS],
            pan_id=0xFFFF if pan_id is None else pan_id,
            extended_pan_id=ExtendedPanId(os.urandom(8))
            if extended_pan_id is None
            else extended_pan_id,
            network_key=t.KeyData(os.urandom(16)),
            reset=False,
        )

        # We want to receive all ZDO callbacks to proxy them back go zipgy
        await self._znp.nvram_write(NwkNvIds.ZDO_DIRECT_CB, t.Bool(True))

        # Reset now so that the changes take effect
        await self._reset()

        await self._znp.request(
            c.AppConfig.BDBStartCommissioning.Req(
                Mode=c.app_config.BDBCommissioningMode.NwkFormation
            ),
            RspStatus=t.Status.SUCCESS,
        )

        # This may take a while because of some sort of background scanning.
        # This can probably be disabled.
        await self._znp.wait_for_response(
            c.ZDO.StateChangeInd.Callback(State=t.DeviceState.StartedAsCoordinator)
        )

        await self._znp.request(
            c.AppConfig.BDBStartCommissioning.Req(
                Mode=c.app_config.BDBCommissioningMode.NwkSteering
            ),
            RspStatus=t.Status.SUCCESS,
        )

        # Create the NV item that keeps track of whether or not we're configured
        osal_create_rsp = await self._znp.request(
            c.Sys.OSALNVItemInit.Req(
                Id=NwkNvIds.HAS_CONFIGURED_ZSTACK3, ItemLen=1, Value=b""
            )
        )

        if osal_create_rsp.Status not in (t.Status.SUCCESS, t.Status.NV_ITEM_UNINIT):
            raise RuntimeError(
                "Could not create HAS_CONFIGURED_ZSTACK3 NV item"
            )  # pragma: no cover

        # Initializing the item won't guarantee that it holds this exact value
        await self._znp.nvram_write(NwkNvIds.HAS_CONFIGURED_ZSTACK3, b"\x55")

    async def _send_zdo_request(
        self, dst_addr, dst_ep, src_ep, cluster, sequence, options, radius, data
    ):
        """
        Zigpy doesn't send ZDO requests via TI's ZDO_* MT commands,
        so it will never receive a reply because ZNP intercepts ZDO replies, never
        sends a DataConfirm, and instead replies with one of its ZDO_* MT responses.

        This method translates the ZDO_* MT response into one zigpy can handle.
        """

        LOGGER.trace(
            "Intercepted a ZDO request: dst_addr=%s, dst_ep=%s, src_ep=%s, "
            "cluster=%s, sequence=%s, options=%s, radius=%s, data=%s",
            dst_addr,
            dst_ep,
            src_ep,
            cluster,
            sequence,
            options,
            radius,
            data,
        )

        assert dst_ep == ZDO_ENDPOINT

        # Deserialize the ZDO request
        zdo_hdr, data = ZDOHeader.deserialize(cluster, data)
        field_names, field_types = ZDO_CLUSTERS[cluster]
        zdo_args, _ = list_deserialize(data, field_types)
        zdo_kwargs = dict(zip(field_names, zdo_args))

        device = self.get_device(nwk=dst_addr.address)

        # Call the converter with the ZDO request's kwargs
        req_factory, rsp_factory, zdo_rsp_factory = ZDO_COMPLEX_CONVERTERS[cluster]
        request = req_factory(dst_addr.address, device, **zdo_kwargs)
        callback = rsp_factory(dst_addr.address)

        LOGGER.debug(
            "Intercepted AP ZDO request %s(%s) and replaced with %s",
            cluster,
            zdo_kwargs,
            request,
        )

        try:
            async with async_timeout.timeout(ZDO_REQUEST_TIMEOUT):
                response = await self._znp.request_callback_rsp(
                    request=request, RspStatus=t.Status.SUCCESS, callback=callback
                )
        except InvalidCommandResponse as e:
            raise DeliveryError(f"Could not send command: {e.response.Status}") from e

        zdo_rsp_cluster, zdo_response_kwargs = zdo_rsp_factory(response)

        self._receive_zdo_message(
            cluster=zdo_rsp_cluster, tsn=sequence, sender=device, **zdo_response_kwargs
        )

        return response.Status, "Request sent successfully"

    async def _send_request(
        self, dst_addr, dst_ep, src_ep, cluster, sequence, options, radius, data
    ):
        if dst_ep == ZDO_ENDPOINT and not (
            cluster == ZDOCmd.Mgmt_Permit_Joining_req
            and dst_addr.mode == t.AddrMode.Broadcast
        ):
            return await self._send_zdo_request(
                dst_addr, dst_ep, src_ep, cluster, sequence, options, radius, data
            )

        request = c.AF.DataRequestExt.Req(
            DstAddrModeAddress=dst_addr,
            DstEndpoint=dst_ep,
            DstPanId=0x0000,
            SrcEndpoint=src_ep,
            ClusterId=cluster,
            TSN=sequence,
            Options=options,
            Radius=radius,
            Data=data,
        )

        if dst_addr.mode == t.AddrMode.Broadcast:
            # We won't always get a data confirmation
            response = await self._znp.request(
                request=request, RspStatus=t.Status.SUCCESS
            )
        else:
            async with async_timeout.timeout(DATA_CONFIRM_TIMEOUT):
                response = await self._znp.request_callback_rsp(
                    request=request,
                    RspStatus=t.Status.SUCCESS,
                    callback=c.AF.DataConfirm.Callback(
                        partial=True, Endpoint=dst_ep, TSN=sequence
                    ),
                )

            LOGGER.debug("Received a data request confirmation: %s", response)

        if response.Status != t.Status.SUCCESS:
            return response.Status, "Invalid response status"

        return response.Status, "Request sent successfully"

    @zigpy.util.retryable_request
    async def request(
        self,
        device,
        profile,
        cluster,
        src_ep,
        dst_ep,
        sequence,
        data,
        expect_reply=True,
        use_ieee=False,
    ):
        tx_options = c.af.TransmitOptions.RouteDiscovery

        # if expect_reply:
        #    tx_options |= c.af.TransmitOptions.APSAck

        if use_ieee:
            destination = t.AddrModeAddress(mode=t.AddrMode.IEEE, address=device.ieee)
        else:
            destination = t.AddrModeAddress(mode=t.AddrMode.NWK, address=device.nwk)

        return await self._send_request(
            dst_addr=destination,
            dst_ep=dst_ep,
            src_ep=src_ep,
            cluster=cluster,
            sequence=sequence,
            options=tx_options,
            radius=30,
            data=data,
        )

    async def broadcast(
        self,
        profile,
        cluster,
        src_ep,
        dst_ep,
        grpid,
        radius,
        sequence,
        data,
        broadcast_address=zigpy.types.BroadcastAddress.RX_ON_WHEN_IDLE,
    ):
        assert grpid == 0

        return await self._send_request(
            dst_addr=t.AddrModeAddress(
                mode=t.AddrMode.Broadcast, address=broadcast_address
            ),
            dst_ep=dst_ep,
            src_ep=src_ep,
            cluster=cluster,
            sequence=sequence,
            options=c.af.TransmitOptions.NONE,
            radius=radius,
            data=data,
        )

    async def mrequest(
        self,
        group_id,
        profile,
        cluster,
        src_ep,
        sequence,
        data,
        *,
        hops=0,
        non_member_radius=3,
    ):
        raise NotImplementedError()  # pragma: no cover

    async def force_remove(self, device) -> None:
        """Forcibly remove device from NCP."""
        leave_rsp = await self._znp.request_callback_rsp(
            request=c.ZDO.MgmtLeaveReq.Req(
                DstAddr=0x0000,  # We handle it
                IEEE=device.ieee,
                RemoveChildren_Rejoin=c.zdo.LeaveOptions.NONE,
            ),
            RspStatus=t.Status.SUCCESS,
            callback=c.ZDO.MgmtLeaveRsp.Callback(Src=0x0000, partial=True),
        )

        assert leave_rsp.Status == t.ZDOStatus.SUCCESS

        # TODO: see what happens when we forcibly remove a device that isn't our child

    async def permit_ncp(self, time_s: int) -> None:
        response = await self._znp.request_callback_rsp(
            request=c.ZDO.MgmtPermitJoinReq.Req(
                AddrMode=t.AddrMode.Broadcast,
                Dst=zigpy.types.BroadcastAddress.ALL_DEVICES,
                Duration=time_s,
                # "This field shall always have a value of 1,
                #  indicating a request to change the
                #  Trust Center policy."
                TCSignificance=1,
            ),
            RspStatus=t.Status.SUCCESS,
            callback=c.ZDO.MgmtPermitJoinRsp.Callback(partial=True),
        )

        if response.Status != t.Status.SUCCESS:
            raise RuntimeError(f"Permit join response failure: {response}")



class InvalidFrame(ValueError):
    pass


class CommandError(Exception):
    pass


class CommandNotRecognized(CommandError):
    pass


class InvalidCommandResponse(CommandError):
    def __init__(self, message, response):
        super().__init__(message)
        self.response = response



"""commands to configure parameters of the device, trust center and BDB subsystem."""

import zigpy_znp.types as t


class TimeoutIndex(t.enum_uint8):
    Seconds_10 = 0x00

    Minutes_2 = 0x01
    Minutes_4 = 0x02
    Minutes_8 = 0x03
    Minutes_16 = 0x04
    Minutes_32 = 0x05
    Minutes_64 = 0x06
    Minutes_128 = 0x07
    Minutes_256 = 0x08
    Minutes_512 = 0x09
    Minutes_1024 = 0x0A
    Minutes_2048 = 0x0B
    Minutes_4096 = 0x0C
    Minutes_8192 = 0x0D
    Minutes_16384 = 0x0E


class CentralizedLinkKeyMode(t.enum_uint8):
    UseDefault = 0x00
    UseProvidedInstallCode = 0x01
    UseProvidedInstallCodeAndFallbackToDefault = 0x02
    UseProvidedAPSLinkKey = 0x03
    UseProvidedAPSLinkKeyAndFallbackToDefault = 0x04


class BDBCommissioningStatus(t.enum_uint8):
    Success = 0x00
    InProgress = 0x01
    NoNetwork = 0x02
    TLTargetFailure = 0x03
    TLNotAaCapable = 0x04
    TLNoScanResponse = 0x05
    TLNotPermitted = 0x06
    TCLKExFailure = 0x07
    FormationFailure = 0x08
    Initialization = 0x00
    FBTargetInProgress = 0x09
    FBInitiatorInProgress = 0x0A
    FBNoIdentifyQueryResponse = 0x0B
    FBBindingTableFull = 0x0C
    NetworkRestored = 0x0D
    Failure = 0x0E


class BDBCommissioningMode(t.enum_flag_uint8):
    NONE = 0

    InitiatorTouchLink = 1 << 0
    NwkSteering = 1 << 1
    NwkFormation = 1 << 2
    FindingBinding = 1 << 3
    Touchlink = 1 << 4
    ParentLost = 1 << 5


class AppConfig(t.CommandsBase, subsystem=t.Subsystem.APPConfig):
    # sets the network frame counter to the value specified in the Frame Counter Value.
    # For projects with multiple instances of frame counter, the message sets the
    # frame counter of the current network
    SetNwkFrameCounter = t.CommandDef(
        t.CommandType.SREQ,
        0xFF,
        req_schema=t.Schema(
            (
                t.Param("FrameCounterValue", t.uint32_t, "network frame counter"),
            )  # XXX: check source for actual size
        ),
        rsp_schema=t.STATUS_SCHEMA,
    )

    # Set the default value used by parent device to expire legacy child devices
    SetDefaultRemoteEndDeviceTimeout = t.CommandDef(
        t.CommandType.SREQ,
        0x01,
        req_schema=t.Schema(
            (
                t.Param(
                    "TimeoutIndex", TimeoutIndex, "0x00 -- 10s otherwise 2^N minutes"
                ),
            )
        ),
        rsp_schema=t.STATUS_SCHEMA,
    )

    # Sets in ZED the timeout value to be send to parent device for child expiring
    SetEndDeviceTimeout = t.CommandDef(
        t.CommandType.SREQ,
        0x02,
        req_schema=t.Schema(
            (
                t.Param(
                    "TimeoutIndex", TimeoutIndex, "0x00 -- 10s otherwise 2^N minutes"
                ),
            )
        ),
        rsp_schema=t.STATUS_SCHEMA,
    )

    # Set the AllowRejoin TC policy
    SetAllowRejoinTCPolicy = t.CommandDef(
        t.CommandType.SREQ,
        0x03,
        req_schema=t.Schema(
            (
                t.Param(
                    "AllowRejoin",
                    t.Bool,
                    "whether or not the Trust center allows rejoins",
                ),
            )
        ),
        rsp_schema=t.STATUS_SCHEMA,
    )

    # Set the commissioning methods to be executed. Initialization of BDB is executed
    # with this call, regardless of its parameters
    BDBStartCommissioning = t.CommandDef(
        t.CommandType.SREQ,
        0x05,
        req_schema=t.Schema(
            (t.Param("Mode", BDBCommissioningMode, "Commissioning mode"),)
        ),
        rsp_schema=t.STATUS_SCHEMA,
    )

    # Set BDB primary or secondary channel masks
    BDBSetChannel = t.CommandDef(
        t.CommandType.SREQ,
        0x08,
        req_schema=t.Schema(
            (
                t.Param("IsPrimary", t.Bool, "True -- is primary channel"),
                t.Param("Channel", t.Channels, "Channel set mask"),
            )
        ),
        rsp_schema=t.STATUS_SCHEMA,
    )

    # Add a preconfigured key (plain key or IC) to Trust Center device
    BDBAddInstallCode = t.CommandDef(
        t.CommandType.SREQ,
        0x04,
        req_schema=t.Schema(
            (
                t.Param(
                    "InstallCodeFormat",
                    t.uint8_t,
                    (
                        "0x01 -- Install code + CRC"
                        "0x02 -- Key derived from install code"
                    ),
                ),
                t.Param("IEEE", t.EUI64, "IEEE address of the joining device"),
                t.Param(
                    "InstallCode", t.Bytes, "16 bytes for derived key, 18 for IC + CRC"
                ),
            )
        ),
        rsp_schema=t.STATUS_SCHEMA,
    )

    # Set the policy flag on Trust Center device to mandate or not the TCLK
    # exchange procedure
    BDBSetTcRequireKeyExchange = t.CommandDef(
        t.CommandType.SREQ,
        0x09,
        req_schema=t.Schema(
            (
                t.Param(
                    "BdbTrustCenterRequireKeyExchange", t.Bool, "Require key exchange"
                ),
            )
        ),
        rsp_schema=t.STATUS_SCHEMA,
    )

    # Sets the policy to mandate or not the usage of an Install Code upon joining
    BDBSetJoinUsesInstallCodeKey = t.CommandDef(
        t.CommandType.SREQ,
        0x06,
        req_schema=t.Schema(
            (t.Param("BdbJoinUsesInstallCodeKey", t.Bool, "Use install code"),)
        ),
        rsp_schema=t.STATUS_SCHEMA,
    )

    # On joining devices, set the default key or an install code to attempt
    # to join the network
    BDBSetActiveDefaultCentralizedKey = t.CommandDef(
        t.CommandType.SREQ,
        0x07,
        req_schema=t.Schema(
            (
                t.Param(
                    "CentralizedLinkKeyModes",
                    CentralizedLinkKeyMode,
                    (
                        "which key will be used when performing association "
                        "to a centralized network"
                    ),
                ),
                t.Param("InstallCode", t.Bytes, "key in any of its formats"),
            )
        ),
        rsp_schema=t.STATUS_SCHEMA,
    )

    # Instruct the ZED to try to rejoin its previews network. Use only in ZED devices
    BDBZedAttemptRecoverNWK = t.CommandDef(
        t.CommandType.SREQ, 0x0A, req_schema=t.Schema(), rsp_schema=t.STATUS_SCHEMA
    )

    # MT_APP_CONFIG Callbacks
    # Callback to receive notifications from BDB process
    BDBCommissioningNotification = t.CommandDef(
        t.CommandType.AREQ,
        0x80,
        rsp_schema=t.Schema(
            (
                t.Param(
                    "Status",
                    BDBCommissioningStatus,
                    "Status of the commissioning mode notified",
                ),
                t.Param(
                    "Mode",
                    BDBCommissioningMode,
                    "Commissioning mode to which status is related",
                ),
                t.Param(
                    "RemainingModes",
                    BDBCommissioningMode,
                    (
                        "Bitmask of the remaining commissioning modes after "
                        "this notification"
                    ),
                ),
            )
        ),
    )
